import numpy as np

from deampy import econ_eval as ce

N = 100

np.random.seed(573)
s_center = np.array([[10000, 0.2], [20000, 0.7], [50000, 1.2]])


s0 = ce.Strategy("s0", s_center[0, 0]+np.random.normal(0, 1000, N),
                 s_center[0, 1]+np.random.normal(0, 0.01, N),
                 color='red')
s1 = ce.Strategy("s1", s_center[1, 0]+np.random.normal(0, 1000, N),
                 s_center[1, 1]+np.random.normal(0, 0.01, N),
                 color='blue')
s2 = ce.Strategy("s2", s_center[2, 0]+np.random.normal(0, 1000, N),
                 s_center[2, 1]+np.random.normal(0, 0.05, N),
                 color='green')

cea = ce.CEA([s0, s1, s2], if_paired=True)
cea.plot_CE_plane()

nmb_paired = ce.CBA([s0, s1, s2],
                    wtp_range=[0, 100000],
                    if_paired=True)
nmb_indp = ce.CBA([s0, s1, s2],
                  wtp_range=[0, 100000],
                  if_paired=False)  # list of frontier strategies as input

# Try NMB_Lines figure - paired CI
nmb_paired.plot_incremental_nmbs(
    title="deltaNMB lines for paired CI",
    x_label="wtp values",
    y_label="NMB values",
    y_axis_multiplier=0.1,
    interval_type='c',
    show_legend=True,
    figure_size=(6, 5))

# Try NMB_Lines figure - paired PI
nmb_paired.plot_incremental_nmbs(
    "deltaNMB lines for paired PI", "wtp values", "NMB values",
    interval_type='p',
    show_legend=True,
    figure_size=(6, 5))

# Try NMB_Lines figure - indp CI
nmb_indp.plot_incremental_nmbs(
    "deltaNMB lines for indp CI", "wtp values", "NMB values",
    interval_type='c',
    show_legend=True,
    figure_size=(6, 5))

# Try NMB_Lines figure - indp PI
nmb_indp.plot_incremental_nmbs(
    "deltaNMB lines for indp PI", "wtp values", "NMB values",
    interval_type='p',
    show_legend=True,
    figure_size=(6, 5))





import deampy.data_structure as df

rows = [
    [0,     0,  10],
    [0,     1,  20],
    [5,     0,  11],
    [5,     1,  12],
    [10,    0,  15],
    [10,    1,  16]
]

df1 = df.DataFrame(rows=rows,
                   list_x_min=[0, 0],
                   list_x_max=[10, 1],
                   list_x_delta=[5, 'int'])

print('print rows:')
print(df1.get_rows())
print('print objects:')
print(df1.get_objs())
print('print %:')
print(df1.get_percents())

for obj in df1.get_objs_gen():
    print(obj)
print('')

print('print: 10 = ', df1.get_value(x_value=[0, 0]))
print('print: 10 = ', df1.get_value(x_value=[1, 0]))
print('print: 20 = ', df1.get_value(x_value=[2, 1]))
print('print: 12 = ', df1.get_value(x_value=[5, 1]))
print('print: 11 = ', df1.get_value(x_value=[9, 0]))
print('print: 16 = ', df1.get_value(x_value=[10, 1]))
print('print: 16 = ', df1.get_value(x_value=[20, 1]))
print('')

rows = [
    [0,     0,  0,  10],
    [0,     0,  1,  11],
    [0,     1,  0,  20],
    [0,     1,  1,  21],
    [5,     0,  0,  30],
    [5,     0,  1,  31],
    [5,     1,  0,  40],
    [5,     1,  1,  41]
]

df2 = df.DataFrame(rows=rows,
                   list_x_min=[0, 0, 0],
                   list_x_max=[5, 1, 1],
                   list_x_delta=[5, 'int', 'int'])

print('print rows:')
print(df2.get_rows())
print('')

print('print: 10 = ', df2.get_value(x_value=[0, 0, 0]))
print('print: 10 = ', df2.get_value(x_value=[1, 0, 0]))
print('print: 20 = ', df2.get_value(x_value=[2, 1, 0]))
print('print: 31 = ', df2.get_value(x_value=[5, 0, 1]))
print('print: 30 = ', df2.get_value(x_value=[9, 0, 0]))
print('print: 41 = ', df2.get_value(x_value=[9, 1, 1]))



import math

import numpy as np

import deampy.data_structure as df
import deampy.random_variats as RVGs

rows = [
    [0, 0, 0, 0.12],
    [0, 0, 1, 0.06],
    [0, 1, 0, 0.17],
    [0, 1, 1, 0.08],
    [5, 0, 0, 0.09],
    [5, 0, 1, 0.25],
    [5, 1, 0, 0.11],
    [5, 1, 1, 0.12],
]

rng = RVGs.RNG(seed=1)
probDf = df.DataFrameWithEmpiricalDist(rows=rows,
                                       list_x_min=[0, 0, 0],
                                       list_x_max=[5, 1, 1],
                                       list_x_delta=[5, 'int', 'int'])
# get a sample
print('Get a sampled index:', probDf.sample_indices(rng=rng))
print('Get a sampled value:', probDf.sample_values(rng=rng))
print('')

# testing to make sure sample by index works
counts = [0]*8
for i in range(5000):
    idx = probDf.sample_indices(rng=rng)
    counts[idx[0]*4 + idx[1]*2 + idx[2]] += 1

print('Testing the sampling by index:')
print(np.array(counts)/sum(counts))
print('')

# testing to make sure sample by value is working
for i in range(5000):
    values = probDf.sample_values(rng=rng)
    i_0 = math.floor(values[0]/5)
    counts[i_0*4 + + values[1]*2 + values[2]] += 1

print('Testing the sampling by value:')
print(np.array(counts)/sum(counts))




import numpy as np

from deampy.parameters import Surge

# 200% increase from the base value which has a default value of 1
surge = Surge(par_max_percent_change=2, par_t0=0, par_t1=2)
# 100% decrease from the base value which has a default value of 0
drop = Surge(par_max_percent_change=-1, par_t0=0, par_t1=2)

ts = np.linspace(0.0, 2.0, 21)

surge_fs = [surge.sample(time=t) for t in ts]
drop_fs = [drop.sample(time=t) for t in ts]

print(ts)
print(surge_fs)
print(drop_fs)



import deampy.discrete_event_sim as Cls
import deampy.support.simulation as sim

# create a simulation event calendar
calendar = Cls.SimulationCalendar()

# create simulation events and add them to the calendar
calendar.add_event(Cls.SimulationEvent(time=2.0, priority=1))
calendar.add_event(Cls.SimulationEvent(time=10.0, priority=2))
calendar.add_event(Cls.SimulationEvent(time=1.1, priority=2))
calendar.add_event(Cls.SimulationEvent(time=1.1, priority=1))

# print the first two events
nextEvent = calendar.get_next_event()
print(nextEvent.time, nextEvent.priority)

nextEvent = calendar.get_next_event()
print(nextEvent.time, nextEvent.priority)


# testing trace
def make_trace():
    calendar = Cls.SimulationCalendar()
    myTrace = sim.Trace(True, 2)
    myTrace._add_message(calendar.time, "Message 1")
    myTrace._add_message(calendar.time, "Message 2")

    myTrace.print_trace(filename='TraceTest2.txt', directory='trace')


make_trace()

#Cls.clear_trace_files('../tests/trace')



from deampy import in_out_functions as OutSupport


myList = [['Col1', 'Col2', 'Col3', 'Col4']]
#myList.append(['s1', 's2', 's3'])
#myList.append(['A', '-100', '1,000'])

for i in range(1, 10):
    myList.append(['row'+str(i), i, 2*i, 3*i])

OutSupport.write_csv(file_name ='myCSV.csv', rows=myList, directory='CSVFolder')



import deampy.random_variats as RVGs

from deampy.optimization.appx_policy_itr import SimModel


class Xto2(SimModel):
    # a simple simulation model that represents x0^2 + x1^2 + noise (where noise is normally distributed)
    def __init__(self, err_sigma):
        """"
        :param err_sigma is the standard deviation of noise term
        """

        SimModel.__init__(self)

        # create a normal distribution to model noise
        self._err = RVGs.Normal(loc=0, scale=err_sigma)

    def get_obj_value(self, x, seed_index=0):
        """ returns one realization from x^2+noise """
        # create a random number generator
        rng = RVGs.RNG(seed=seed_index)

        return (x[0]+1)*(x[0]+1) + x[1]*x[1] + self._err.sample(rng)


class Xto2Constrained(SimModel):
    # a simple simulation model that represents x0^2 + x1^2 + noise (where noise is normally distributed)
    # and x1 should be greater than 1
    def __init__(self, err_sigma, penalty):
        """"
        :param err_sigma is the standard deviation of noise term
        """

        SimModel.__init__(self)
        # create a normal distribution to model noise
        self._err = RVGs.Normal(loc=0, scale=err_sigma)
        # penalty
        self._penalty = penalty

    def get_obj_value(self, x, seed_index=0):
        """ returns one realization from x^2+noise """

        # create a random number generator
        rng = RVGs.RNG(seed=seed_index)

        accum_penalty = 0       # accumulated penalty

        # test the feasibility
        if x[1] < 1:
            accum_penalty += self._penalty * pow(x[1] - 1, 2)
            x[1] = 1

        return (x[0]+1)*(x[0]+1) + x[1]*x[1] + self._err.sample(rng) + accum_penalty



import numpy as np

import deampy.plots.histogram as cls

obs = np.random.normal(4, 3, 1000)

cls.plot_histogram(
    data=obs,
    title='Histogram',
    x_label='Values',
    y_label='Counts',
    color='g',
    bin_width=1,
    x_range=[-5, 20],
    y_range=[0, 140],
    legend='Number of patients')

obs_sets = [
    np.random.normal(4, 3, 1000),
    np.random.normal(8, 3, 1000)
]

cls.plot_histograms(
    data_sets=obs_sets,
    title='Two histograms',
    x_label='Values',
    y_label='Counts',
    legends=['H 1', 'H 2'],
    bin_width=1,
    x_range=[-10, 20],
    #y_range=[0, 100],
    color_codes=['blue', 'green'],
    transparency=0.6
)


import os
import string

import deampy.format_functions as F
from deampy.support.misc_functions import *


def output_figure(plt, filename=None, dpi=300, bbox_inches='tight'):
    """
    :param plt: reference to the plot
    :param filename: filename to save this figure as (e.g. 'figure.png') (if None, the figure will be displayed)
    :param dpi: dpi of the figure
    """
    # output
    if filename is None:
        plt.show()
    else:
        if filename[0] == '/':
            filename = filename[1:]
        # get directory
        directory_path = os.path.dirname(filename)

        # create the directory if does not exist
        if directory_path != '':
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)

        try:
            plt.savefig(proper_file_name(filename), dpi=dpi, bbox_inches=bbox_inches)
        except ValueError:
            raise ValueError("Error in saving figure '{}'. "
                             "Ensure that the filename is valid and "
                             "that the folder where the figure should be saved exists.".format(filename))


def calculate_ticks(l, u, delta):
    values = []
    v = l
    while v <= u:
        values.append(v)
        v += delta

    return values


def format_x_axis(ax, min_x, max_x, delta_x, buffer=0, form=None, deci=None):

    # get x ticks
    xs = ax.get_xticks()

    if min_x is None:
        min_x = xs[0]
    if max_x is None:
        max_x = xs[-1]
    if delta_x is not None:
        xs = calculate_ticks(min_x, max_x, delta_x)

    # format x-axis
    ax.set_xticks(xs)
    if deci is not None:
        ax.set_xticklabels([F.format_number(x, deci=deci, format=form) for x in xs])

    ax.set_xlim([min_x-buffer, max_x+buffer])


def add_labels_to_panels(axarr, x_coord=-0.2, y_coord=1, font_size=8):
    """
    adds A), B), etc. labels to panels
    :param axarr: (array) of panels
    :param x_coord: (float) increase to move labels right
    :param y_coord: (float) increase to move labels up
    :param font_size: (float) font size of labels
    """
    axs = axarr.flat
    for n, ax in enumerate(axs):
        ax.text(x_coord, y_coord,
                string.ascii_uppercase[n] + ')',
                transform=ax.transAxes,
                size=font_size,
                weight='bold')



import numpy as np

import deampy.regression as Reg
from deampy.regression import ExpRegression as exp

# parameters of an exponential function
par1 = (3, 2, 0.5)  # 3 + 2*exp(0.5*x)
par2 = (2, 0.5)     # 2*exp(0.5*x)

# x values between
x = np.linspace(0, 4, 50)

# y values
y1 = exp.exp_func(x, *par1)
y2 = exp.exp_func_c0_zero(x, *par2)

# adding noice to y values
y1n = y1 + 0.2*np.random.normal(size=len(x))
y2n = y2 + 0.2*np.random.normal(size=len(x))

# fitting an exponential function
exp_reg_1 = Reg.ExpRegression(x, y1n, if_c0_zero=False)
exp_reg_2 = Reg.ExpRegression(x, y2n, if_c0_zero=True)

# display the fit
exp_reg_1.plot_fit()
exp_reg_2.plot_fit()

# print the estimated coefficients
print('\ncoefficients of y1 model :', exp_reg_1.get_coeffs(), 'v.s', par1)
print('prediction vs actual y1 at x=2 :', exp_reg_1.get_predicted_y(2), 'v.s', exp.exp_func(2, *par1))

print('\ncoefficients of y2 model :', exp_reg_2.get_coeffs(), 'v.s', par2)
print('prediction vs actual y2 at x=2: ', exp_reg_2.get_predicted_y(2), 'v.s', exp.exp_func_c0_zero(2, *par2))

# testing on with p0
x = [112.25, 728.185, 879.07, 937.03, 963.485, 982.22, 996.4, 1001.005,	1006.87, 1170.445]
y = [1.07, 17.68, 28.53, 35.675, 41.59, 48.05, 57.47, 59.605, 66.565, 206.935]

exp_reg_3 = Reg .ExpRegression(x, y, if_c0_zero=True)
exp_reg_3_with_p0 = Reg .ExpRegression(x, y, if_c0_zero=True, p0=[0.5, 0.005])
exp_reg_3.plot_fit()
exp_reg_3_with_p0.plot_fit()
print('\ncoefficients of y3 model (without initial value)', exp_reg_3.get_coeffs())
print('coefficients of y3 model (with initial value)', exp_reg_3_with_p0.get_coeffs())



import numpy as np

from deampy.regression import PolynomialQFunction

N = 1000
sigma = 0
l2_reg = 0.01
forgetting_factor = 1 + 0 * 0.95
degree = 2

# coefficients of [1, x1, x2, x1^2, x1.x2, x2^2] if the categorical variable i 0
COEFFS_C0 = np.array([-1, 1, 2, 3, -2, 4])
# coefficients of [1, x1, x2, x1^2, x1.x2, x2^2] if the categorical variable is 1
COEFFS_C1 = np.array([1, -1, 3, -2, 3, 1])
# coefficients of [-1, 2] for f(i) = -1 + 2i
COEFFS_I = np.array([-1, 2])


def f(x):
    """
    :param x: (list) continuous variables
    :return: f(x)
    """
    return COEFFS_C0 @ np.array([1, x[0], x[1], pow(x[0], 2), x[0] * x[1], pow(x[1], 2)])


def g(x, i):
    """
    :param x: (list) continuous variables
    :param i: categorical variable
    :return: g(x, i)
    """
    return (COEFFS_C0 + i * COEFFS_C1) @ np.array([1, x[0], x[1], pow(x[0], 2), x[0] * x[1], pow(x[1], 2)])


def h(i):
    return COEFFS_I @ np.array([1, i])


np.random.seed(seed=1)

# generate X (to store observations on the continuous variable)
# and I (to store observations ont the categorical variable)
X = []
I = []
for n in range(N):

    # 2 random numbers over [-1, 1] for (x1, x2)
    x = 2*np.random.sample(2)-1

    # 0 or 1 for the categorical variable
    i = 0 if np.random.random_sample() < 0.5 else 1

    X.append(x)
    I.append(i)

# find f(x)
fs = [f(x) + np.random.normal(0, sigma) for x in X]
gs = [g(x, i) + np.random.normal(0, sigma) for x, i in zip(X, I)]
hs = [h(i) for i in I]

# fit a Q-function with continuous variables only
q_cont = PolynomialQFunction(degree=degree, l2_penalty=l2_reg)
for i in range(N):
    q_cont.update(f=fs[i], values_of_continuous_features=X[i])
print('\nQ-function with continuous variables: ')
print('Coeffs: ', q_cont.get_coeffs(), 'vs.', COEFFS_C0)
print('f([1, -1]) should be 7 = ', q_cont.f(values_of_continuous_features=[1, -1]))

# fit a Q-function with a categorical variable
q_cat = PolynomialQFunction(degree=0, l2_penalty=l2_reg)
for i in range(N):
    q_cat.update(f=hs[i], values_of_indicator_features=I[i])
print('\nQ-function with categorical variables: ')
print('Coeffs: ', q_cat.get_coeffs(), 'vs.', COEFFS_I)
print('h([1]) should be 1) = ', q_cat.f(values_of_indicator_features=1))

# fit a Q-function with continuous and categorical variables
q_cont_cat = PolynomialQFunction(degree=degree, l2_penalty=l2_reg)
for i in range(N):
    q_cont_cat.update(f=gs[i], values_of_continuous_features=X[i], values_of_indicator_features=I[i])
print('\nQ-function with continuous and categorical variables: ')
print('Coeffs: ', q_cont_cat.get_coeffs(), 'vs.', np.append(COEFFS_C0, COEFFS_C1))
print('f([1, -1, 0] should be 7) = ', q_cont_cat.f(values_of_continuous_features=[1, -1], values_of_indicator_features=0))
print('f([1, -1, 1] should be 0) = ', q_cont_cat.f(values_of_continuous_features=[1, -1], values_of_indicator_features=1))



import compare as C
from deampy.optimization.appx_policy_itr import ApproximatePolicyIteration
from deampy.optimization.learning_exploring_rules import *
from model import Model

N_ITRS = 5000
B = 50
BETA = 0.5
Q_FUNC_DEGREE = 2
L2_PENALTY = 0.01

sim_model = Model(cost_sigma=C.COST_SIGMA,
                  action_cost=C.ACTION_COST)

api = ApproximatePolicyIteration(sim_model=sim_model,
                                 num_of_actions=1,
                                 learning_rule=Harmonic(b=B),
                                 exploration_rule=EpsilonGreedy(beta=BETA),
                                 q_function_degree=Q_FUNC_DEGREE,
                                 l2_penalty=L2_PENALTY)

api.minimize(n_iterations=N_ITRS)

api.export_results(csv_file='iterations.csv')
api.plot_iterations(moving_ave_window=int(N_ITRS / 20), fig_size=(5, 6),
                    n_last_iterations_to_ave=N_ITRS/5)

C.compare(q_function_degree=Q_FUNC_DEGREE)




from deampy.plots import sample_paths as Path

# a sample path with initial size = 1
path1 = Path.PrevalenceSamplePath('Path 1', initial_size=1, sim_rep=1)
# record the observations
path1.record_increment(1.2, 2)
path1.record_increment(2.3, -1)
path1.record_increment(5.1, 0)

# second sample path with initial size = 1
path2 = Path.PrevalenceSamplePath('Path 2', initial_size=0, sim_rep=1)
# record the observations
path2.record_increment(0.5, 4)
path2.record_increment(1.8, -2)
path2.record_increment(5.5, 1)

# third sample path with initial size = 1
path3 = Path.PrevalencePathBatchUpdate(
    'Path 3', 0, times_of_changes=[1.5, 2, 5], increments=[2, -1, 0], sim_rep=1)
path4 = Path.PrevalencePathBatchUpdate(
    'Path 4', 0, times_of_changes=[0.5, 4, 4.5], increments=[2.5, 1, 1], sim_rep=1)

# stats
print(path1.stat.get_mean())


# plot path 1 only
Path.plot_sample_path(
    sample_path=path1,
    title='Plotting a single sample path',
    x_label='time',
    y_label='observed value',
    legend='Path 1',
    color_code='r',
    connect='step')

# plot path 3 only
Path.plot_sample_path(
    sample_path=path3,
    title='Plotting a single sample path that is updated in batch',
    x_label='time',
    y_label='observed value',
    legend='Path 3',
    color_code='r')

# plot both paths
Path.plot_sample_paths(
    sample_paths=[path1, path2],
    title='Plot two sample paths with different color',
    x_label='time',
    y_label='observed value',
    legends=['Path 1', 'Path 2'],
    transparency=0.75)

Path.plot_sample_paths(
    sample_paths=[path1, path2],
    title='Plot 2 sample paths with the same color',
    x_label='time',
    y_label='observed value',
    legends='Path',
    transparency=0.5,
    common_color_code='g')

Path.plot_sets_of_sample_paths(
    sets_of_sample_paths=[[path1, path2], [path3, path4]],
    title='Plot 2 sets of sample paths',
    x_label='time',
    y_label='observed value',
    legends=['Set 1', 'Set 2'],
    transparency=0.5,
    color_codes=['g', 'b'])



import string
import warnings

import matplotlib.cm as cm
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import scipy.stats as stat
from numpy import exp, power, average
from numpy.random import RandomState
from scipy.stats import sem

import deampy.format_functions as F
import deampy.in_out_functions as IO
from deampy.plots.econ_eval_plots import old_add_curves_to_ax, add_curves_to_ax
from deampy.statistics import SummaryStat
from deampy.support.econ_eval_support import *
from deampy.support.misc_classes import *
from deampy.support.misc_functions import convert_lnl_to_prob

# warnings.filterwarnings("always")
NUM_OF_BOOTSTRAPS = 1000  # number of bootstrap samples to calculate confidence intervals for ICER
LEGEND_FONT_SIZE = 7

CE_FRONTIER_COLOR = 'k'
CE_FRONTIER_TRANSPARENCY = 0.6
CE_FRONTIER_LINE_WIDTH = 2
CE_EDGE_COLOR = 'w'

CEAC_LINE_WIDTH = 0.75
CEAF_LINE_WIDTH = 2.5

# for constrained optimization and NMB
NMB_LINE_WIDTH = 0.75
NMB_INTERVAL_TRANSPARENCY = 0.5
NMB_FRONTIER_LINE_WIDTH = 2.5

FRONTIER_LABEL_SHIFT_X = -0.01 # shift labels to right or left (proportional to the length of the x_axis)
FRONTIER_LABEL_SHIFT_Y = 0.01 # shift labels to right or left (proportional to the length of the x_axis)


def pv_single_payment(payment, discount_rate, discount_period, discount_continuously=False):
    """ calculates the present value of a single future payment
    :param payment: payment to calculate the present value for
    :param discount_rate: discount rate
    :param discount_period: number of periods to discount the payment
    :param discount_continuously: set to True to discount continuously
    :return: payment/(1+discount_rate)^discount_period for discrete discounting
             payment * exp(-discounted_rate*discount_period) for continuous discounting """

    # error checking
    if discount_continuously:
        pass
    else:
        assert type(discount_period) is int, "discount_period should be an integer number."
    if discount_rate < 0 or discount_rate > 1:
        raise ValueError("discount_rate should be a number between 0 and 1.")
    if discount_period <= 0:
        raise ValueError("discount_period should be greater than 0.")

    # calculate the present value
    if discount_continuously:
        return payment * exp(-discount_rate * discount_period)
    else:
        return payment * power(1 + discount_rate, -discount_period)


def pv_continuous_payment(payment, discount_rate, discount_period):
    """ calculates the present value of a future continuous payment (discounted continuously)
    :param payment: payment to calculate the present value for
    :param discount_rate: discount rate
    :param discount_period: (tuple) in the form of (l, u) specifying the period where
                             the continuous payment is received
    :return: payment * (exp(-discount_rate*l - exp(-discount_rate*u))/discount_rate
    """
    assert type(discount_period) is tuple, "discount_period should be a tuple (l, u)."
    if discount_rate < 0 or discount_rate > 1:
        raise ValueError("discount_rate should be a number between 0 and 1.")

    if discount_rate == 0:
        return payment * (discount_period[1] - discount_period[0])
    else:
        return payment/discount_rate * \
               (exp(-discount_rate*discount_period[0])
                - exp(-discount_rate*discount_period[1]))


def equivalent_annual_value(present_value, discount_rate, discount_period):
    """  calculates the equivalent annual value of a present value
    :param present_value:
    :param discount_rate: discount rate (per period)
    :param discount_period: number of periods to discount the payment
    :return: discount_rate*present_value/(1-pow(1+discount_rate, -discount_period))
    """

    # error checking
    assert type(discount_period) is int, "discount_period should be an integer number."
    if discount_rate < 0 or discount_rate > 1:
        raise ValueError("discount_rate should be a number between 0 and 1.")
    if discount_period < 0:
        raise ValueError("discount_period cannot be less than 0.")

    # calculate the equivalent annual value
    return discount_rate*present_value/(1-power(1+discount_rate, -discount_period))


class Strategy:
    def __init__(self, name, cost_obs, effect_obs, color=None, marker='o', label=None, short_label=None):
        """
        :param name: name of the strategy
        :param cost_obs: list or numpy.array of cost observations
        :param effect_obs: list or numpy.array of effect observations
        :param color: (string) color code
                (https://www.webucator.com/blog/2015/03/python-color-constants-module/)
        :param marker: (string) marker code
                (https://matplotlib.org/3.1.1/api/markers_api.html)
        :param label: (string) label to show on the legend (if None, name is used)
        :param short_label: (string) label to show on the center of the probability clouds
            or on the curves of of NMBs (if None, label is used)
        """

        assert color is None or type(color) is str, "color argument should be a string."

        self.idx = 0        # index of the strategy
        self.name = name
        self.color = color
        self.marker = marker

        self.label = name if label is None else label
        self.shortLabel = self.label if short_label is None else short_label

        self.ifDominated = False
        self.switchingWTP = 0
        self.switchingWTPInterval = []

        self.costObs = assert_np_list(cost_obs,
                                      error_message='cost_obs should be a list or a np.array')
        self.dCostObs = None    # (list) cost observations with respect to base
        self.incCostObs = None  # (list) incremental cost observations
        self.cost = None        # summary statistics for cost
        self.dCost = None       # summary statistics for cost with respect to base
        self.incCost = None     # summary statistics for incremental cost

        self.effectObs = assert_np_list(effect_obs,
                                        error_message='effect_obs should be a list or a np.array')
        self.dEffectObs = None      # (list) effect observations with respect to base
        self.incEffectObs = None    # (list) incremental effect observations
        self.effect = None          # summary statistics for effect
        self.dEffect = None         # summary statistics for effect with respect to base
        self.incEffect = None       # summary statistics for incremental effect

        self.cer = None         # cost-effectiveness ratio with respect to base
        self.icer = None        # icer summary statistics
        self.eIncNMB = None     # summary statistics for expected incremental net health benefit
                                # integrated over a wtp distribution

        self.cost = Stat.SummaryStat(name='Cost of '+name, data=self.costObs)
        self.effect = Stat.SummaryStat(name='Effect of '+name, data=self.effectObs)

    def reset(self):
        """ set class attributes that will be calculated later to None """

        self.ifDominated = False
        self.dCostObs = None
        self.incCostObs = None
        self.dCost = None
        self.incCost = None

        self.dEffectObs = None
        self.incEffectObs = None
        self.dEffect = None
        self.incEffect = None

        self.cer = None
        self.icer = None

    def get_cost_err_interval(self, interval_type, alpha=0.05, multiplier=1):
        """
        :param interval_type: (string) 'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param alpha: significance level
        :param multiplier: to multiply the estimate and the interval by the provided value
        :return: list [err_l, err_u] for the lower and upper error length
                of confidence or prediction intervals of cost observations.
                NOTE: [err_l, err_u] = [mean - L, mean + U], where [L, U] is the confidence or prediction interval

        """
        interval = self.cost.get_interval(interval_type, alpha, multiplier)
        return [self.cost.get_mean()*multiplier - interval[0],
                interval[1] - self.cost.get_mean()*multiplier]

    def get_effect_err_interval(self, interval_type, alpha, multiplier=1):
        """
        :param interval_type: (string) 'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param alpha: significance level
        :param multiplier: to multiply the estimate and the interval by the provided value
        :return: list [err_l, err_u] for the lower and upper error length
                of confidence or prediction intervals of effect observations.
                NOTE: [err_l, err_u] = [mean - L, mean + U], where [L, U] is the confidence or prediction interval

        """
        interval = self.effect.get_interval(interval_type, alpha, multiplier)
        return [self.effect.get_mean()*multiplier - interval[0],
                interval[1] - self.effect.get_mean()*multiplier]


class _EconEval:
    """ super class for cost-effective analysis (CEA) and cost-benefit analysis (CBA) """

    def __init__(self, strategies, if_paired, health_measure='u', if_reset_strategies=False):
        """
        :param strategies: the list of strategies (assumes that the first strategy represents the "base" strategy)
        :param if_paired: set to true to indicate that the strategies are paired
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        :param if_reset_strategies: set to True if the cost and effect with respect to
            base, incremental cost and effect, ICER, and CER of strategies should be recalculated.
        """

        if health_measure not in ['u', 'd']:
            raise ValueError("health_measure can be either 'u' (for utility) or 'd' (for disutility).")

        self.strategies = strategies  # list of strategies
        # assign the index of each strategy
        for i, s in enumerate(strategies):
            s.idx = i
            if if_reset_strategies:
                s.reset()

        self._n = len(strategies)  # number of strategies
        self._ifPaired = if_paired  # if cost and effect outcomes are paired across strategies
        self._healthMeasure = health_measure  # utility of disutility
        self._u_or_d = 1 if health_measure == 'u' else -1

        # assign colors to strategies
        self.__assign_colors()

    def __assign_colors(self):
        """ assignes color to each strategy if colors are not provided """

        # decide about the color of each curve
        rainbow_colors = cm.rainbow(np.linspace(0, 1, self._n))
        for i, s in enumerate(self.strategies):
            if s.color is None:
                s.color = rainbow_colors[i]

    def _find_shifted_strategies(self):
        """ find shifted strategies.
        In calculating the change in effect, it accounts for whether QALY or DALY is used.
        """

        # shift all strategies such that the base strategy (first in the list) lies on the origin
        # if observations are paired across strategies
        if self._ifPaired:

            for i, s in enumerate(self.strategies):

                s.dCostObs = s.costObs - self.strategies[0].costObs
                s.dCost = Stat.DifferenceStatPaired(name='Cost with respect to base',
                                                    x=s.costObs,
                                                    y_ref=self.strategies[0].costObs)
                # if health measure is utility
                if self._healthMeasure == 'u':
                    s.dEffectObs = s.effectObs - self.strategies[0].effectObs
                    s.dEffect = Stat.DifferenceStatPaired(name='Effect with respect to base',
                                                          x=s.effectObs,
                                                          y_ref=self.strategies[0].effectObs)

                else:  # if health measure is disutility
                    s.dEffectObs = self.strategies[0].effectObs - s.effectObs
                    s.dEffect = Stat.DifferenceStatPaired(name='Effect with respect to base',
                                                          x=self.strategies[0].effectObs,
                                                          y_ref=s.effectObs)

                # cost-effectiveness ratio of non-base strategies
                if i > 0:
                    s.cer = ICER_Paired(name='Cost-effectiveness ratio of ' + s.name,
                                        costs_new=s.costObs,
                                        effects_new=s.effectObs,
                                        costs_base=self.strategies[0].costObs,
                                        effects_base=self.strategies[0].effectObs,
                                        health_measure=self._healthMeasure)

        else:  # if not paired
            # get average cost and effect of the base strategy
            base_ave_cost = self.strategies[0].cost.get_mean()
            base_ave_effect = self.strategies[0].effect.get_mean()

            for i, s in enumerate(self.strategies):
                s.dCostObs = s.costObs - base_ave_cost
                s.dCost = Stat.DifferenceStatIndp(name='Cost with respect to base',
                                                  x=s.costObs,
                                                  y_ref=self.strategies[0].costObs)
                if self._healthMeasure == 'u':
                    s.dEffectObs = s.effectObs - base_ave_effect
                    s.dEffect = Stat.DifferenceStatIndp(name='Effect with respect to base',
                                                        x=s.effectObs,
                                                        y_ref=self.strategies[0].effectObs)

                else:  # if health measure is disutility
                    s.dEffectObs = base_ave_effect - s.effectObs
                    s.dEffect = Stat.DifferenceStatIndp(name='Effect with respect to base',
                                                        x=self.strategies[0].effectObs,
                                                        y_ref=s.effectObs)

                # cost-effectiveness ratio of non-base strategies
                if i > 0:
                    s.cer = ICER_Indp(name='Cost-effectiveness ratio of ' + s.name,
                                      costs_new=s.costObs,
                                      effects_new=s.effectObs,
                                      costs_base=self.strategies[0].costObs,
                                      effects_base=self.strategies[0].effectObs,
                                      health_measure=self._healthMeasure)


class CEA(_EconEval):
    """ super class for cost-effective analysis (CEA) and cost-benefit analysis (CBA) """

    def __init__(self, strategies, if_paired, health_measure='u', if_reset_strategies=False):
        """
        :param strategies: the list of strategies (assumes that the first strategy represents the "base" strategy)
        :param if_paired: set to true to indicate that the strategies are paired
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        :param if_reset_strategies: set to True if the cost and effect with respect to
            base, incremental cost and effect, ICER, and CER of strategies should be recalculated.
        """

        _EconEval.__init__(self, strategies=strategies,
                           if_paired=if_paired,
                           health_measure=health_measure,
                           if_reset_strategies=if_reset_strategies)

        self._strategies_on_frontier = []   # list of strategies on the frontier
        self._ifFrontierIsCalculated = False  # CE frontier is not calculated yet
        self._ifPairwiseCEAsAreCalculated = False
        self._pairwise_ceas = []  # list of list to cea's

        # shift the strategies
        self._find_shifted_strategies()

        # find the cost-effectiveness frontier
        self.__find_frontier()

    def get_strategies_on_frontier(self):
        """
        :return: strategies on the frontier sorted in the increasing order of effect
        """

        if not self._ifFrontierIsCalculated:
            self.__find_frontier()

        return self._strategies_on_frontier
    
    def get_strategies_not_on_frontier(self):
        """
        :return: strategies not on the frontier 
        """
        
        if not self._ifFrontierIsCalculated:
            self.__find_frontier()
        
        return [s for s in self.strategies if s.ifDominated]

    def build_CE_table(self,
                       interval_type='c',
                       alpha=0.05, method='bootstrap', num_bootstrap_samples=1000, rng=None,
                       prior_range=None, num_wtp_thresholds=1000,
                       cost_digits=0, effect_digits=2, icer_digits=1,
                       cost_multiplier=1, effect_multiplier=1,
                       file_name='CETable.csv', directory=''):
        """
        :param interval_type: (string) 'n' for no interval
                                       'c' or 'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param alpha: significance level, a value from [0, 1]
        :param method: (string) method to calculate confidence interval ('bootstrap' or 'Bayesian')
        :param num_bootstrap_samples: number of bootstrap samples when 'bootstrap' method is selected
        :param rng: random number generator to generate empirical bootstrap samples
        :param num_wtp_thresholds: (int) number of willingness-to-pay thresholds to evaluate posterior
            when 'Bayesian' approach is selected
        :param prior_range: (tuple) in form of (l, u) for the prior range of willingness-to-pay
            threshold that makes NMB zero (if prior is not provided [0, 4 * ICER] will be used
        :param cost_digits: digits to round cost estimates to
        :param effect_digits: digits to round effect estimate to
        :param icer_digits: digits to round ICER estimates to
        :param cost_multiplier: set to 1/1000 or 1/100000 to represent cost in terms of
                thousands or hundred thousands unit
        :param effect_multiplier: set to 1/1000 or 1/100000 to represent effect in terms of
                thousands or hundred thousands unit
        :param file_name: address and file name where the CEA results should be saved to
        :param directory: directory (relative to the current root) where the files should be located
            for example use 'Example' to create and save the csv file under the folder Example
        """

        # find the frontier if not calculated already
        if not self._ifFrontierIsCalculated:
            self.__find_frontier()

        table = [['Strategy', 'Cost', 'Effect', 'Incremental Cost', 'Incremental Effect', 'ICER']]
        # sort strategies in increasing order of cost
        self.strategies.sort(key=get_d_cost)

        for i, s in enumerate(self.strategies):
            row=[]
            # strategy name
            row.append(s.name)
            # strategy cost
            row.append(s.cost.get_formatted_mean_and_interval(interval_type=interval_type,
                                                              alpha=alpha,
                                                              deci=cost_digits,
                                                              form=',',
                                                              multiplier=cost_multiplier))
            # strategy effect
            row.append(s.effect.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                alpha=alpha,
                                                                deci=effect_digits,
                                                                form=',',
                                                                multiplier=effect_multiplier))

            # strategy incremental cost
            if s.incCost is None:
                row.append('-')
            else:
                row.append(s.incCost.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                     alpha=alpha,
                                                                     deci=cost_digits,
                                                                     form=',',
                                                                     multiplier=cost_multiplier))
            # strategy incremental effect
            if s.incEffect is None:
                row.append('-')
            else:
                row.append(s.incEffect.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                       alpha=alpha,
                                                                       deci=effect_digits,
                                                                       form=',',
                                                                       multiplier=effect_multiplier))

            # ICER
            if s.ifDominated:
                row.append('Dominated')
            elif s.icer is not None:
                row.append(s.icer.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                  alpha=alpha,
                                                                  method=method,
                                                                  num_bootstrap_samples=num_bootstrap_samples,
                                                                  rng=rng,
                                                                  prior_range=prior_range,
                                                                  num_wtp_thresholds=num_wtp_thresholds,
                                                                  deci=icer_digits,
                                                                  form=',',
                                                                  multiplier=1))
            else:
                row.append('-')

            table.append(row)

        IO.write_csv(file_name=file_name, directory=directory, rows=table, delimiter=',')

        # sort strategies back
        self.strategies.sort(key=get_index)

    def get_dCost_dEffect_cer(self,
                              interval_type='n',
                              alpha=0.05,
                              cost_digits=0, effect_digits=2, icer_digits=1,
                              cost_multiplier=1, effect_multiplier=1):
        """
        :param interval_type: (string) 'n' for no interval,
                                       'c' for confidence interval,
                                       'p' for percentile interval
        :param alpha: significance level
        :param cost_digits: digits to round cost estimates to
        :param effect_digits: digits to round effect estimate to
        :param icer_digits: digits to round ICER estimates to
        :param cost_multiplier: set to 1/1000 or 1/100,000 to represent cost in terms of
                thousands or hundred thousands unit
        :param effect_multiplier: set to 1/1000 or 1/100,000 to represent effect in terms of
                thousands or hundred thousands unit
        :return: a dictionary of additional cost, additional effect, and cost-effectiveness ratio for
                all strategies with respect to the Base strategy
        """

        dictionary_results = {}

        for s in [s for s in self.strategies if s.idx > 0]:

            d_cost_text = s.dCost.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                  alpha=alpha,
                                                                  deci=cost_digits,
                                                                  form=',',
                                                                  multiplier=cost_multiplier)
            d_effect_text = s.dEffect.get_formatted_mean_and_interval(interval_type=interval_type,
                                                                      alpha=alpha,
                                                                      deci=effect_digits,
                                                                      form=',',
                                                                      multiplier=effect_multiplier)
            cer_text = s.cer.get_formatted_mean_and_interval(interval_type=interval_type,
                                                             alpha=alpha,
                                                             deci=icer_digits,
                                                             form=',',
                                                             multiplier=1)
            # add to the dictionary
            dictionary_results[s.name] = [d_cost_text, d_effect_text, cer_text]

        return dictionary_results

    def add_ce_plane_to_ax(self, ax,
                           x_range=None, y_range=None,
                           add_clouds=True, show_legend=True,
                           center_s=50, cloud_s=10, transparency=0.1,
                           cost_multiplier=1, effect_multiplier=1,
                           cost_decimals=None, effect_decimals=None,
                           interval_type=None, significance_level=0.05, interval_transparency=0.5,
                           legend_loc_code=0):
        """
        adds a cost-effectiveness plane to the provided ax
        :param ax: axis
        :param x_range: (tuple) range of x-axis
        :param y_range: (tuple) range of y-axis
        :param add_clouds: (bool) if to add the probability clouds
        :param show_legend: (bool) if to show the legend
        :param center_s: (float) the size of the dot showing (x,y) of a strategy
        :param cloud_s: (float) the size of dots building the probability clouds
        :param transparency: (float) the transparency of dots building the probability clouds
        :param cost_multiplier: (float) to multiply the cost values
        :param effect_multiplier: (float) to multiply the effect values
        :param cost_decimals: (int) to round the labels of cost axis
        :param effect_decimals: (int) to round the labels of the effect axis
        :param interval_type: (string) None to not display the intervals,
                                       'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param significance_level: (float) significance level for the interval
        :param interval_transparency: (float) the transparency of intervals
        :param legend_loc_code: (int) legend location code
            https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
        """

        # find the frontier (x, y)'s
        frontier_d_effect = []
        frontier_d_costs = []
        for s in self.get_strategies_on_frontier():
            frontier_d_effect.append(s.dEffect.get_mean()*effect_multiplier)
            frontier_d_costs.append(s.dCost.get_mean()*cost_multiplier)

        # add all strategies
        for s in self.strategies:
            # the mean change in effect and cost
            ax.scatter(s.dEffect.get_mean()*effect_multiplier, s.dCost.get_mean()*cost_multiplier,
                       color=s.color,  # color
                       alpha=1,  # transparency
                       marker=s.marker,  # markers
                       s=center_s,  # marker size
                       label=s.label,  # label to show in the legend
                       zorder=2,
                       edgecolors=CE_EDGE_COLOR
                       )
            # add error bars
            if interval_type is not None:
                x_mean = s.dEffect.get_mean()*effect_multiplier
                y_mean = s.dCost.get_mean()*cost_multiplier

                x_interval = s.dEffect.get_interval(
                    interval_type=interval_type, alpha=significance_level, multiplier=effect_multiplier)
                y_interval = s.dCost.get_interval(
                    interval_type=interval_type, alpha=significance_level, multiplier=cost_multiplier)

                ax.errorbar(x_mean, y_mean,
                            xerr=[[x_mean-x_interval[0]], [x_interval[1]-x_mean]],
                            yerr=[[y_mean-y_interval[0]], [y_interval[1]-y_mean]],
                            fmt='none', color=s.color, linewidth=1, alpha=interval_transparency)

        # add the frontier line
        if len(self.get_strategies_on_frontier()) > 1:
            ax.plot(frontier_d_effect, frontier_d_costs,
                    color=CE_FRONTIER_COLOR,  # color
                    alpha=CE_FRONTIER_TRANSPARENCY,  # transparency
                    linewidth=CE_FRONTIER_LINE_WIDTH,  # line width
                    zorder=3,
                    label='Frontier',  # label to show in the legend
                    )

        if show_legend:
            ax.legend(fontsize='7.5', loc=legend_loc_code)

        # and the clouds
        if add_clouds:
            # add all strategies
            for s in self.strategies:
                ax.scatter(s.dEffectObs * effect_multiplier, s.dCostObs * cost_multiplier,
                           color=s.color,  # color of dots
                           marker=s.marker, # marker
                           alpha=transparency,  # transparency of dots
                           s=cloud_s,  # size of dots
                           zorder=1
                           )

        ax.set_xlim(x_range)  # x-axis range
        ax.set_ylim(y_range)  # y-axis range

        # format x-axis
        if effect_decimals is not None:
            vals_x = ax.get_xticks()
            ax.set_xticks(vals_x)
            ax.set_xticklabels(['{:,.{prec}f}'.format(x, prec=effect_decimals) for x in vals_x])

        # format y-axis
        if cost_decimals is not None:
            vals_y = ax.get_yticks()
            ax.set_yticks(vals_y)
            ax.set_yticklabels(['{:,.{prec}f}'.format(x, prec=cost_decimals) for x in vals_y])

        ax.axhline(y=0, c='k', linestyle='--', linewidth=0.5)
        ax.axvline(x=0, c='k', linestyle='--', linewidth=0.5)

    def plot_CE_plane(self,
                      title='Cost-Effectiveness Analysis',
                      x_label='Additional Health',
                      y_label='Additional Cost',
                      x_range=None, y_range=None,
                      add_clouds=False, fig_size=(5, 5),
                      show_legend=True,
                      center_s=75, cloud_s=25, transparency=0.1,
                      cost_multiplier=1, effect_multiplier=1,
                      cost_digits=0, effect_digits=1,
                      interval_type=None, significance_level=0.05, interval_transparency=0.5,
                      file_name=None
                      ):
        ''' plots a cost-effectiveness plane
        :param title: (string) title of the figure
        :param x_label: (string) label of x-axis
        :param y_label: (string) label of y-axis
        :param x_range: (tuple) (minimum value, maximum value) of the y-axis
        :param y_range: (tuple) (minimum value, maximum value) of the y-axis
        :param add_clouds: (boolean) set to True to show the projection clouds
        :param fig_size: (tuple) (width, height) of the figure
        :param show_legend: (boolean) set to True to show the legends
        :param center_s: (float) size of dots that show the mean cost and health of each strategy
        :param cloud_s: (float) size of dots that form the clouds
        :param transparency: (float between 0 and 1) transparency of dots that form the clouds
        :param cost_multiplier: (float) set to 1/1000 or 1/100000 to represent cost in terms of
                thousands or hundred thousands unit
        :param effect_multiplier: (float) set to 1/1000 or 1/100000 to represent effect in terms of
                thousands or hundred thousands unit
        :param cost_digits: (int) number of digits to round cost labels to
        :param effect_digits: (int) number of digits to round effect labels to
        :param interval_type: (string) None to not display the intervals,
                               'c' for t-based confidence interval,
                               'cb' for bootstrap confidence interval, and
                               'p' for percentile interval
        :param significance_level: (float) significance level for the interval
        :param interval_transparency: (float) the transparency of intervals
        :param file_name: (string) file name to save the figure as
        :return:
        '''

        fig, ax = plt.subplots(figsize=fig_size)

        ax.set_title(title)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        # add the cost-effectiveness plane
        self.add_ce_plane_to_ax(ax=ax,
                                x_range=x_range, y_range=y_range,
                                add_clouds=add_clouds,
                                show_legend=show_legend,
                                center_s=center_s, cloud_s=cloud_s, transparency=transparency,
                                cost_multiplier=cost_multiplier, effect_multiplier=effect_multiplier,
                                cost_decimals=cost_digits, effect_decimals=effect_digits,
                                interval_type=interval_type, significance_level=significance_level,
                                interval_transparency=interval_transparency)

        fig.tight_layout()

        if file_name is None or file_name == '':
            fig.show()
        else:
            fig.savefig(file_name, dpi=300)

    def create_pairwise_ceas(self):
        """
        creates a list of list for pairwise cost-effectiveness analysis
        For example for strategies ['Base', 'A', 'B']:
        [
            ['Base wr Base',    'A wr Base',    'B wr Base'],
            ['Base wr A',       'A wr A',       'B wr A'],
            ['B wr B',          'A wr B',       'B wr B'],
        ]
        """

        # create CEA's for all pairs
        self._pairwise_ceas = []
        for s_base in self.strategies:
            list_ceas = []
            for s_new in self.strategies:#[1:]:

                # if the base and the new strategies are the same
                if s_base.name == s_new.name:
                    list_ceas.append(None)
                else:
                    list_ceas.append(CEA(strategies=[s_base, s_new],
                                         if_paired=self._ifPaired,
                                         health_measure=self._healthMeasure,
                                         if_reset_strategies=True)
                                     )
            self._pairwise_ceas.append(list_ceas)

        self._ifPairwiseCEAsAreCalculated = True

        # since the relative performance of strategies
        # (relative costs and relative effects) have changed.
        self._ifFrontierIsCalculated = False

    def print_pairwise_cea(self, interval_type='n',
                           alpha=0.05,
                           cost_digits=0, effect_digits=2, icer_digits=1,
                           cost_multiplier=1, effect_multiplier=1,
                           directory='Pairwise_CEA'):
        """
        :param interval_type: (string) 'n' for no interval,
                                       'c' for confidence interval,
                                       'p' for percentile interval
        :param alpha: significance level
        :param cost_digits: digits to round cost estimates to
        :param effect_digits: digits to round effect estimate to
        :param icer_digits: digits to round ICER estimates to
        :param cost_multiplier: set to 1/1000 or 1/100,000 to represent cost in terms of
                thousands or hundred thousands unit
        :param effect_multiplier: set to 1/1000 or 1/100,000 to represent effect in terms of
                thousands or hundred thousands unit
        :param directory: directory (relative to the current root) where the files should be located
            for example use 'Example' to create and save the csv file under the folder Example
        """

        # create the pair-wise cost-effectiveness analyses
        if not self._ifPairwiseCEAsAreCalculated:
            self.create_pairwise_ceas()

        # save the CEA tables
        for row_of_ceas in self._pairwise_ceas:
            for cea in row_of_ceas:
                if cea is not None:
                    name = cea.strategies[1].name + ' to ' + cea.strategies[0].name
                    cea.build_CE_table(interval_type=interval_type,
                                       alpha=alpha,
                                       cost_digits=cost_digits,
                                       effect_digits=effect_digits,
                                       icer_digits=icer_digits,
                                       cost_multiplier=cost_multiplier,
                                       effect_multiplier=effect_multiplier,
                                       file_name=name+'.csv',
                                       directory=directory)

    def plot_pairwise_ceas(self,
                           figure_size=None, font_size=6,
                           show_subplot_labels=False,
                           effect_label='', cost_label='',
                           center_s=50, cloud_s=25, transparency=0.2,
                           x_range=None, y_range=None,
                           cost_multiplier=1, effect_multiplier=1,
                           column_titles=None, row_titles=None,
                           file_name='pairwise_CEA.png'):

        # identify which CEA is valid
        # (i.e. comparing strategies that are on the frontier)
        # valid comparisons are marked with '*'
        valid_comparison = []
        for i in range(self._n):
            valid_comparison.append(['']*self._n)
        on_frontier = self.get_strategies_on_frontier()
        for idx in range(len(on_frontier)-1):
            i = on_frontier[idx].idx
            j = on_frontier[idx+1].idx
            valid_comparison[i][j] = '*'

        # set default properties of the figure
        plt.rc('font', size=font_size) # fontsize of texts
        plt.rc('axes', titlesize=font_size)  # fontsize of the figure title
        plt.rc('axes', titleweight='semibold')  # fontweight of the figure title

        # plot each panel
        f, axarr = plt.subplots(nrows=self._n, ncols=self._n,
                                sharex=True, sharey=True, figsize=figure_size)

        Y_LABEL_COORD_X = -0.05     # increase to move the A-B-C labels to right
        abc_idx = 0
        for i in range(self._n):
            for j in range(self._n):
                # get the current axis
                ax = axarr[i, j]

                # add the A-B-C label if needed
                if show_subplot_labels:
                    ax.text(Y_LABEL_COORD_X - 0.05, 1.05,
                            string.ascii_uppercase[abc_idx] + ')',
                            transform=ax.transAxes,
                            size=font_size + 1, weight='bold')

                # add titles for the figures_national in the first row
                if i == 0:
                    if column_titles is None:
                        ax.set_title(self.strategies[j].name)
                    else:
                        ax.set_title(column_titles[j])

                # add y_labels for the figures_national in the first column
                if j == 0:
                    if row_titles is None:
                        ax.set_ylabel(self.strategies[i].name, fontweight='bold')
                    else:
                        ax.set_ylabel(row_titles[i], fontweight='bold')

                # specify ranges of x- and y-axis
                if x_range is not None:
                    ax.set_xlim(x_range)
                if y_range is not None:
                    ax.set_ylim(y_range)

                # CEA of these 2 strategies
                cea = CEA(strategies=[self.strategies[i], self.strategies[j]],
                          if_paired=self._ifPaired,
                          health_measure=self._healthMeasure,
                          if_reset_strategies=True)

                # add the CE figure to this axis
                cea.add_ce_plane_to_ax(ax=ax, show_legend=False,
                                       center_s=center_s,
                                       cloud_s=cloud_s,
                                       transparency=transparency,
                                       cost_multiplier=cost_multiplier,
                                       effect_multiplier=effect_multiplier)

                # add ICER to the figure
                # could be 'Dominated', 'Cost-Saving' or 'estimated ICER'
                text = ''
                if i != j and cea.strategies[1].ifDominated:
                    text = 'Dominated'
                elif cea.strategies[1].dCost.get_mean() < 0 and cea.strategies[1].dEffect.get_mean() > 0:
                    text = 'Cost-Saving' + valid_comparison[i][j]
                elif cea.strategies[1].icer is not None:
                    text = F.format_number(cea.strategies[1].icer.get_ICER(), deci=1, format='$') + valid_comparison[i][j]
                # add the text of the ICER to the figure
                ax.text(0.95, 0.95, text, transform=ax.transAxes, fontsize=6,
                        va='top', ha='right')

                abc_idx += 1

        # add the common effect and cost labels
        f.text(0.55, 0, effect_label, ha='center', va='center', fontweight='bold')
        f.text(0.99, 0.5, cost_label, va='center', rotation=-90, fontweight='bold')

        # f.show()
        f.savefig(file_name, bbox_inches='tight', dpi=300)

        # since the relative performance of strategies 
        # (relative costs and relative effects) have changed.
        self._ifFrontierIsCalculated = False

    def __find_frontier(self):

        # apply criteria 1 (strict dominance)
        # if a strategy i yields less health than strategy j but costs more, it is dominated
        # sort by effect with respect to base
        self.strategies.sort(key=get_d_effect)
        for i in range(self._n):
            for j in range(i+1, self._n):
                if self.strategies[i].dCost.get_mean() >= self.strategies[j].dCost.get_mean():
                    self.strategies[i].ifDominated = True
                    break;

        # select all non-dominated strategies
        select_strategies = [s for s in self.strategies if not s.ifDominated]

        # apply criteria 1 (strict dominance)
        # if a strategy i costs more than strategy j but yields less health, it is dominated
        # sort strategies by cost with respect to the base
        select_strategies.sort(key=get_d_cost, reverse=True)
        for i in range(len(select_strategies)):
            for j in range(i + 1, len(select_strategies)):
                if select_strategies[i].dEffect.get_mean() <= select_strategies[j].dEffect.get_mean():
                    select_strategies[i].ifDominated = True
                    break;

        # apply criteria 2 (weak dominance)
        # select all non-dominated strategies
        select_strategies = [s for s in self.strategies if not s.ifDominated]

        for i in range(len(select_strategies)):
            for j in range(i+1, len(select_strategies)):
                # cost and effect of strategy i
                d_cost_i = select_strategies[i].dCost.get_mean()
                d_effect_i = select_strategies[i].dEffect.get_mean()
                # cost and effect of strategy j
                d_cost_j = select_strategies[j].dCost.get_mean()
                d_effect_j = select_strategies[j].dEffect.get_mean()
                # vector connecting strategy i to j
                v_i_to_j = np.array([d_effect_j - d_effect_i, d_cost_j - d_cost_i])

                # find strategies with dEffect between i and j
                s_between_i_and_j = []
                for s in select_strategies:
                    if d_effect_i < s.dEffect.get_mean() < d_effect_j:
                        s_between_i_and_j.append(s)

                # if the dEffect of no strategy is between the effects of strategies i and j
                if len(s_between_i_and_j) == 0:
                    continue  # to the next j
                else:

                    for inner_s in s_between_i_and_j:
                        # vector from i to inner_s
                        v_i_to_inner = np.array([inner_s.dEffect.get_mean() - d_effect_i,
                                                 inner_s.dCost.get_mean() - d_cost_i])

                        # cross products of vector i to j and the vectors i to the inner point
                        cross_product = v_i_to_j[0] * v_i_to_inner[1] - v_i_to_j[1] * v_i_to_inner[0]

                        # if cross_product > 0 the point is above the line
                        # (because the point are sorted vertically)
                        # ref: How to tell whether a point is to the right or left side of a line
                        # https://stackoverflow.com/questions/1560492
                        if cross_product > 0:
                            inner_s.ifDominated = True

        # sort strategies by effect with respect to the base
        self.strategies.sort(key=get_d_effect)

        # find strategies on the frontier
        self._strategies_on_frontier = [s for s in self.strategies if not s.ifDominated]

        # sort back
        self.strategies.sort(key=get_index)

        # frontier is calculated
        self._ifFrontierIsCalculated = True

        # calcualte the incremental outcomes
        self.__calculate_incremental_outcomes()

    def __calculate_incremental_outcomes(self):

        if self._ifPaired:

            for i, s in enumerate(self._strategies_on_frontier):
                if i > 0:
                    s_before = self._strategies_on_frontier[i-1]

                    s.incCostObs = s.costObs - s_before.costObs
                    s.incCost = Stat.DifferenceStatPaired(name='Incremental cost',
                                                          x=s.costObs,
                                                          y_ref=s_before.costObs)
                    # if health measure is utility
                    if self._healthMeasure == 'u':
                        s.incEffectObs = s.effectObs - self._strategies_on_frontier[i-1].effectObs
                        s.incEffect = Stat.DifferenceStatPaired(name='Effect with respect to base',
                                                                x=s.effectObs,
                                                                y_ref=s_before.effectObs)

                    else:  # if health measure is disutility
                        s.incEffectObs = self._strategies_on_frontier[i-1].effectObs - s.effectObs
                        s.incEffect = Stat.DifferenceStatPaired(name='Effect with respect to base',
                                                                x=s_before.effectObs,
                                                                y_ref=s.effectObs)
                    # ICER
                    s.icer = ICER_Paired(name='ICER of {} relative to {}'.format(s.name, s_before.name),
                                         costs_new=s.costObs,
                                         effects_new=s.effectObs,
                                         costs_base=s_before.costObs,
                                         effects_base=s_before.effectObs,
                                         health_measure=self._healthMeasure)

        else:  # if not paired

            for i, s in enumerate(self._strategies_on_frontier):

                if i > 0:
                    s_before = self._strategies_on_frontier[i - 1]

                    # get average cost and effect of the strategy i - 1
                    ave_cost_i_1 =s_before.cost.get_mean()
                    ave_effect_i_1 = s_before.effect.get_mean()

                    s.incCostObs = s.costObs - ave_cost_i_1
                    s.incCost = Stat.DifferenceStatIndp(name='Cost with respect to base',
                                                        x=s.costObs,
                                                        y_ref=s_before.costObs)
                    if self._healthMeasure == 'u':
                        s.incEffectObs = s.effectObs - ave_effect_i_1
                        s.incEffect = Stat.DifferenceStatIndp(name='Effect with respect to base',
                                                              x=s.effectObs,
                                                              y_ref=s_before.effectObs)

                    else:  # if health measure is disutility
                        s.incEffectObs = ave_effect_i_1 - s.effectObs
                        s.incEffect = Stat.DifferenceStatIndp(name='Effect with respect to base',
                                                              x=s_before.effectObs,
                                                              y_ref=s.effectObs)

                    # ICER
                    s.icer = ICER_Indp(name='ICER of {} relative to {}'.format(s.name, s_before.name),
                                       costs_new=s.costObs,
                                       effects_new=s.effectObs,
                                       costs_base=s_before.costObs,
                                       effects_base=s_before.effectObs,
                                       health_measure=self._healthMeasure)


class CBA(_EconEval):
    """ class for doing cost-benefit analysis """

    def __init__(self, strategies, wtp_range, if_paired, health_measure='u', n_of_wtp_values=200):
        """
        :param strategies: the list of strategies (assumes that the first strategy represents the "base" strategy)
        :param wtp_range: ([l, u]) range of willingness-to-pay values over which the NMB analysis should be done
        :param if_paired: indicate whether the strategies are paired
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        :param n_of_wtp_values: number of willingness-to-pay values to construct net monetary benefit curves
        """
        _EconEval.__init__(self, strategies=strategies,
                           if_paired=if_paired,
                           health_measure=health_measure)

        self.inmbCurves = []  # list of incremental NMB curves with respect to the base
        self.acceptabilityCurves = []  # the list of acceptability curves
        self.expectedLossCurves = []  # the list of expected loss curves
        self.evpi = None

        # use net monetary benefit for utility by default
        if health_measure == 'u':
            self.utility = inmb_u
        else:
            self.utility = inmb_d

        # wtp values (includes the specified minimum and maximum wtp value)
        self.wtpValues = np.linspace(wtp_range[0], wtp_range[1],
                                     num=n_of_wtp_values, endpoint=True)

        # index of strategy with the highest expected net-monetary benefit over the wtp range
        self.idxHighestExpNMB = []
        # index of strategy with the lowest expected loss over the wtp range
        self.idxLowestExpLoss = []

    def build_inmb_curves(self, interval_type='n'):
        """
        prepares the information needed to plot the incremental net-monetary benefit
        with respect to the first strategy (base)
        :param interval_type: (string) 'n' for no interval,
                                       'c' for confidence interval,
                                       'p' for percentile interval):
        """

        self.inmbCurves = []  # list of incremental NMB curves

        # create the NMB curves
        for s in self.strategies:

            if self._ifPaired:
                # create a paired NMB object
                inmb = INMB_Paired(name=s.name,
                                   costs_new=s.costObs,
                                   effects_new=s.effectObs,
                                   costs_base=self.strategies[0].costObs,
                                   effects_base=self.strategies[0].effectObs,
                                   health_measure=self._healthMeasure)

            else:
                # create an independent NMB object
                inmb = INMB_Indp(name=s.name,
                                 costs_new=s.costObs,
                                 effects_new=s.effectObs,
                                 costs_base=self.strategies[0].costObs,
                                 effects_base=self.strategies[0].effectObs,
                                 health_measure=self._healthMeasure)

            # make a NMB curve
            self.inmbCurves.append(INMBCurve(label=s.label,
                                             short_label=s.shortLabel,
                                             color=s.color,
                                             wtp_values=self.wtpValues,
                                             inmb_stat=inmb,
                                             interval_type=interval_type)
                                   )

        self.idxHighestExpNMB = update_curves_with_highest_values(
            wtp_values=self.wtpValues, curves=self.inmbCurves)

    def build_acceptability_curves(self):
        """
        prepares the information needed to plot the cost-effectiveness acceptability curves
        """

        if not self._ifPaired:
            raise ValueError('Calculating the acceptability curves when outcomes are not paired'
                             'across strategies is not implemented.')

        # initialize acceptability curves
        self.acceptabilityCurves = []
        for s in self.strategies:
            self.acceptabilityCurves.append(
                AcceptabilityCurve(label=s.label,
                                   short_label=s.shortLabel,
                                   color=s.color))

        n_obs = len(self.strategies[0].costObs)

        # for each WTP value, calculate the number of times that
        # each strategy has the highest NMB value
        for w in self.wtpValues:

            # number of times that each strategy is optimal
            count_maximum = np.zeros(self._n)

            for obs_idx in range(n_obs):

                # find which strategy has the maximum:
                max_nmb = float('-inf')
                max_s_i = 0  # index of the optimal strategy for this observation
                for s_i, s in enumerate(self.strategies):
                    d_effect = (s.effectObs[obs_idx] - self.strategies[0].effectObs[obs_idx]) * self._u_or_d
                    d_cost = s.costObs[obs_idx] - self.strategies[0].costObs[obs_idx]
                    nmb = w * d_effect - d_cost
                    if nmb > max_nmb:
                        max_nmb = nmb
                        max_s_i = s_i

                count_maximum[max_s_i] += 1

            # calculate probabilities that each strategy has been optimal
            prob_maximum = count_maximum / n_obs

            for i in range(self._n):
                self.acceptabilityCurves[i].xs.append(w)
                self.acceptabilityCurves[i].ys.append(prob_maximum[i])

        if len(self.idxHighestExpNMB) == 0:
            self.idxHighestExpNMB = update_curves_with_highest_values(
                wtp_values=self.wtpValues, curves=self.inmbCurves)

        # find the optimal strategy for each wtp value
        for wtp_idx, wtp in enumerate(self.wtpValues):
            opt_idx = self.idxHighestExpNMB[wtp_idx]
            self.acceptabilityCurves[opt_idx].maxXs.append(wtp)
            self.acceptabilityCurves[opt_idx].maxYs.append(
                self.acceptabilityCurves[opt_idx].ys[wtp_idx])

        for c in self.acceptabilityCurves:
            c.convert_lists_to_arrays()

    def build_expected_loss_curves(self):
        """
        prepares the information needed to plot the expected loss curves
        """

        if not self._ifPaired:
            raise ValueError('Calculating the expected loss curves when outcomes are not paired'
                             'across strategies is not implemented.')

        # initialize expected loss curves
        self.expectedLossCurves = []
        for s in self.strategies:
            self.expectedLossCurves.append(
                ExpectedLossCurve(label=s.label,
                                  short_label=s.shortLabel,
                                  color=s.color))

        n_obs = len(self.strategies[0].costObs)

        # for each WTP value, calculate the number of times that
        # each strategy has the highest NMB value
        for i, w in enumerate(self.wtpValues):

            mean_max_nmb = 0
            for obs_idx in range(n_obs):

                # find which strategy has the maximum nmb:
                max_nmb = float('-inf')
                max_s_i = 0  # index of the optimal strategy for this observation
                for s_i, s in enumerate(self.strategies):
                    d_effect = (s.effectObs[obs_idx] - self.strategies[0].effectObs[obs_idx]) * self._u_or_d
                    d_cost = s.costObs[obs_idx] - self.strategies[0].costObs[obs_idx]
                    nmb = w * d_effect - d_cost
                    if nmb > max_nmb:
                        max_nmb = nmb
                        max_s_i = s_i

                mean_max_nmb += max_nmb

            # calculate probabilities that each strategy has been optimal
            mean_max_nmb = mean_max_nmb / n_obs

            for s_i in range(self._n):
                self.expectedLossCurves[s_i].xs.append(w)
                self.expectedLossCurves[s_i].ys.append(mean_max_nmb - self.inmbCurves[s_i].ys[i])

        if len(self.idxLowestExpLoss) == 0:
            self.idxLowestExpLoss = update_curves_with_lowest_values(
                wtp_values=self.wtpValues, curves=self.expectedLossCurves)

    def find_optimal_switching_wtp_values(self, interval_type='n', deci=0):

        w_stars = []  # wtp values to switch between strategies
        w_star_intervals = [] # confidence or projection intervals of optimal wtp values
        s_stars = []  # indices of optimal strategies between wtp values
        s_star_names = []   # names of optimal strategies

        # working wtp value
        w = self.wtpValues[0]

        # find the optimal strategy at wtp = 0
        max_nmb = float('-inf')
        s_star = 0
        for s in self.strategies:
            u = self.utility(d_effect=s.dEffect.get_mean(),
                             d_cost=s.dCost.get_mean())
            u_value = u(w)
            if u_value > max_nmb:
                max_nmb = u_value
                s_star = s.idx

        # record the information of the optimal strategy at wtp = 0
        w_stars.append(w)
        w_star_intervals.append([w, w])
        s_stars.append(s_star)
        s_star_names.append(self.strategies[s_star].name)

        # find the optimal switching wtp values
        while w <= self.wtpValues[-1] and len(s_stars) < len(self.strategies):

            # find the intersect of the current strategy with other
            w_min = float('inf')
            for s in self.strategies:
                if s.idx not in s_stars:

                    line = Line(x1=self.strategies[s_star].dEffect.get_mean(),
                                x2=s.dEffect.get_mean(),
                                y1=self.strategies[s_star].dCost.get_mean(),
                                y2=s.dCost.get_mean())
                    w_star = line.slope

                    if w_star is not math.nan and w_star >= w:
                        if w_star < w_min:
                            w_min = w_star
                            s_star = s.idx

            w = w_min
            if w != float('inf'):
                w_stars.append(w)
                s_stars.append(s_star)
                s_star_names.append(self.strategies[s_star].name)

        # find and store the confidence or projection intervals of
        # switching wtp values
        for i, s_index in enumerate(s_stars[0:-1]):
            next_s_index = s_stars[i+1]
            if self._ifPaired:
                inmb = INMB_Paired(
                    name='',
                    costs_new=self.strategies[next_s_index].costObs,
                    effects_new=self.strategies[next_s_index].effectObs,
                    costs_base=self.strategies[s_index].costObs,
                    effects_base=self.strategies[s_index].effectObs,
                    health_measure=self._healthMeasure
                )
            else:
                inmb = INMB_Indp(
                    name='',
                    costs_new=self.strategies[next_s_index].costObs,
                    effects_new=self.strategies[next_s_index].effectObs,
                    costs_base=self.strategies[s_index].costObs,
                    effects_base=self.strategies[s_index].effectObs,
                    health_measure=self._healthMeasure
                )

            w_star, interval = inmb.get_switch_wtp_and_interval(
                wtp_range=[self.wtpValues[0], self.wtpValues[-1]],
                interval_type=interval_type
            )
            w_star_intervals.append(interval)

        # update status of strategies
        i = 0
        for s in self.strategies:
            if s.idx in s_stars:
                s.ifDominated = False
                s.switchingWTP = w_stars[i]
                s.switchingWTPInterval = w_star_intervals[i]
                i += 1
            else:
                s.ifDominated = True

        # populate the results to report back
        result = [['Strategy', 'ID', 'WTP', 'Interval']]
        for i in range(len(s_stars)):
            result.append(
                [
                    s_star_names[i],
                    s_stars[i],
                    F.format_number(w_stars[i], deci=deci, format=','),
                    F.format_interval(w_star_intervals[i], deci=deci, format=',')
                ]
            )

        return result

    def calculate_evpi_curve(self):
        """ calculates the expected value of perfect information (EVPI) curve """

        self.evpi = []
        n_of_sims = len(self.strategies[0].dCostObs)

        # for all budget value
        for wtp in self.wtpValues:

            # find the highest achievable NMB under perfect information
            max_nmbs = []
            for i in range(n_of_sims):
                # find costs and effects of strategies for the ith monte carlo simulation run
                costs = [s.dCostObs[i] for s in self.strategies]
                effects = [s.dEffectObs[i] for s in self.strategies]

                # find the maximum effect
                max_nmb = float('-inf')
                for c, e in zip(costs, effects):
                    nmb = wtp * e - c
                    if nmb > max_nmb:
                        max_nmb = nmb
                max_nmbs.append(max_nmb)

            self.evpi.append(average(max_nmbs))

        # curve
        self.inmbCurves.append(EVPI(xs=self.wtpValues, ys=self.evpi, label='EVPI', color='k'))

    def plot_incremental_nmbs(self,
                              title='Incremental Net Monetary Benefit',
                              x_label='Willingness-To-Pay Threshold',
                              y_label='Expected Incremental Net Monetary Benefit',
                              show_evpi=False,
                              y_range=None,
                              y_axis_multiplier=1,
                              y_axis_decimal=0,
                              interval_type='c',
                              delta_wtp=None,
                              transparency_lines=0.5,
                              transparency_intervals=0.2,
                              show_legend=True,
                              show_labels_on_frontier=False,
                              figure_size=(5, 5),
                              file_name=None):
        """
        plots the incremental net-monetary benefit of each strategy
                with respect to the base (the first strategy)
        :param title: title
        :param x_label: x-axis label
        :param y_label: y-axis label
        :param show_evpi: (bool) if to show EVPI curve
        :param y_range: (list) range of y-axis
        :param y_axis_multiplier: (float) multiplier to scale the y-axis
            (e.g. 0.001 for thousands)
        :param y_axis_decimal: (float) decimal of y_axis numbers
        :param interval_type: (string) 'n' for no interval,
                                       'c' for confidence interval,
                                       'p' for percentile interval
        :param delta_wtp: distance between the labels of WTP values shown on the x-axis
        :param transparency_lines: transparency of net monetary benefit lines (0.0 transparent through 1.0 opaque)
        :param transparency_intervals: transparency of intervals (0.0 transparent through 1.0 opaque)
        :param show_legend: set true to show legend
        :param show_labels_on_frontier: set true to show strategy labels on frontier
        :param figure_size: (tuple) size of the figure (e.g. (2, 3)
        :param file_name: (string) filename to save the figure as
        """

        # make incremental NMB curves
        self.build_inmb_curves(interval_type=interval_type)

        if show_evpi:
            self.calculate_evpi_curve()

        # initialize plot
        fig, ax = plt.subplots(figsize=figure_size)

        # add the incremental NMB curves
        add_curves_to_ax(ax=ax, curves=self.inmbCurves,
                         x_values=self.wtpValues,
                         title=title, x_label=x_label,
                         y_label=y_label, y_range=y_range, x_delta=delta_wtp,
                         y_axis_multiplier=y_axis_multiplier,
                         y_axis_decimal=y_axis_decimal,
                         transparency_lines=transparency_lines,
                         transparency_intervals=transparency_intervals,
                         show_legend=show_legend,
                         show_labels_on_frontier=show_labels_on_frontier)

        fig.tight_layout()

        if file_name is None:
            fig.show()
        else:
            fig.savefig(file_name, dpi=300)

    def add_inmb_curves_to_ax(self, ax,
                              title='Incremental Net Monetary Benefit',
                              x_label='Willingness-To-Pay Threshold',
                              y_label='Expected Incremental Net Monetary Benefit',
                              show_evpi=False,
                              y_range=None,
                              y_axis_multiplier=1,
                              y_axis_decimal=None,
                              delta_wtp=None,
                              interval_type='n',
                              show_legend=True,
                              show_labels_on_frontier=False):

        # make incremental NMB curves
        self.build_inmb_curves(interval_type=interval_type)

        if show_evpi:
            self.calculate_evpi_curve()

        add_curves_to_ax(ax=ax, curves=self.inmbCurves, x_values=self.wtpValues,
                         title=title, x_label=x_label,
                         y_label=y_label, y_range=y_range, x_delta=delta_wtp,
                         y_axis_decimal=y_axis_decimal,
                         y_axis_multiplier=y_axis_multiplier,
                         transparency_lines=1,
                         transparency_intervals=NMB_INTERVAL_TRANSPARENCY,
                         show_legend=show_legend,
                         show_labels_on_frontier=show_labels_on_frontier,
                         show_frontier=True,
                         curve_line_width=NMB_LINE_WIDTH,
                         frontier_line_width=NMB_FRONTIER_LINE_WIDTH,
                         if_format_y_numbers=True if y_axis_decimal is not None else False,
                         frontier_label_shift_x=FRONTIER_LABEL_SHIFT_X,
                         frontier_label_shift_y=FRONTIER_LABEL_SHIFT_Y)

    def plot_acceptability_curves(self,
                                  title=None,
                                  x_label='Willingness-To-Pay Threshold',
                                  y_label='Probability of Being the Optimal Strategy',
                                  y_range=None,
                                  delta_wtp=None,
                                  show_legend=True, fig_size=(5, 5),
                                  legends=None,
                                  file_name=None):
        """
        plots the acceptability curves
        :param title: title
        :param x_label: x-axis label
        :param y_label: y-axis label
        :param y_range: (list) range of y-axis
        :param delta_wtp: distance between the labels of WTP values shown on the x-axis
        :param show_legend: set true to show legend
        :param legends: (list) of legends to display on the figure
        :param fig_size: (tuple) size of the figure (e.g. (2, 3)
        :param file_name: file name
        """

        # make the NMB curves
        self.build_inmb_curves(interval_type='n')

        # make the acceptability curves
        self.build_acceptability_curves()

        # initialize plot
        fig, ax = plt.subplots(figsize=fig_size)

        # add the incremental NMB curves
        self.add_acceptability_curves_to_ax(ax=ax,
                                            wtp_delta=delta_wtp,
                                            y_range=y_range,
                                            show_legend=show_legend,
                                            legends=legends)

        ax.set_title(title)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        if file_name is None:
            fig.show()
        else:
            fig.savefig(file_name, bbox_inches='tight', dpi=300)

    def add_acceptability_curves_to_ax(
            self, ax, wtp_delta=None, y_range=None, show_legend=True, legends=None):
        """
        adds the acceptability curves to the provided ax
        :param ax: axis
        :param wtp_delta: (float) distance between ticks on x-axis
        :param y_range: (tuple) range of y-axis
        :param show_legend: (bool) if to show the legend
        :param legends: (list of strings) texts for legends
        """

        if len(self.inmbCurves) == 0:
            self.build_inmb_curves(interval_type='n')

        if len(self.acceptabilityCurves) == 0:
            self.build_acceptability_curves()

        old_add_curves_to_ax(ax=ax,
                             curves=self.acceptabilityCurves,
                             legends=legends,
                             x_range=[self.wtpValues[0], self.wtpValues[-1]],
                             x_delta=wtp_delta,
                             y_range=y_range, show_legend=show_legend,
                             line_width=CEAC_LINE_WIDTH, opt_line_width=CEAF_LINE_WIDTH,
                             legend_font_size=LEGEND_FONT_SIZE, opt='max')

    def add_expected_loss_curves_to_ax(
            self, ax, wtp_delta=None, y_range=None, show_legend=True, legends=None,
            y_axis_multiplier=1, y_axis_decimal=None):
        """
        adds the acceptability curves to the provided ax
        :param ax: axis
        :param wtp_delta: (float) distance between ticks on x-axis
        :param y_range: (tuple) range of y-axis
        :param show_legend: (bool) if to show the legend
        :param legends: (list of strings) texts for legends
        """

        if len(self.inmbCurves) == 0:
            self.build_inmb_curves(interval_type='n')

        if len(self.expectedLossCurves) == 0:
            self.build_expected_loss_curves()

        old_add_curves_to_ax(ax=ax,
                             curves=self.expectedLossCurves,
                             legends=legends,
                             x_range=[self.wtpValues[0], self.wtpValues[-1]],
                             x_delta=wtp_delta,
                             y_range=y_range, show_legend=show_legend,
                             line_width=CEAC_LINE_WIDTH, opt_line_width=CEAF_LINE_WIDTH,
                             legend_font_size=LEGEND_FONT_SIZE,
                             y_axis_multiplier=y_axis_multiplier, y_axis_decimal=y_axis_decimal,
                             if_y_axis_prob=False, opt='min')

    def get_w_starts(self):

        w_stars = []
        s_stars = []
        s_star_names = []

        w = self.wtpValues[0]

        # at initial w
        max_nmb = float('-inf')
        s_star = 0
        for s in self.strategies:
            u = self.utility(d_effect=s.dEffect.get_mean(),
                             d_cost=s.dCost.get_mean())
            u_value = u(w)
            if u_value > max_nmb:
                max_nmb = u_value
                s_star = s.idx

        w_stars.append(w)
        s_stars.append(s_star)
        s_star_names.append(self.strategies[s_star].name)

        while w <= self.wtpValues[-1] and len(s_stars) < len(self.strategies):

            # find the intersect of the current strategy with other
            w_min = float('inf')
            for s in self.strategies:
                if s.idx not in s_stars:

                    w_star = find_intersecting_wtp(
                        w0=w,
                        u_new=self.utility(d_effect=s.dEffect.get_mean(),
                                           d_cost=s.dCost.get_mean()),
                        u_base=self.utility(d_effect=self.strategies[s_star].dEffect.get_mean(),
                                            d_cost=self.strategies[s_star].dCost.get_mean()))

                    if w_star is not None:
                        if w_star < w_min:
                            w_min = w_star
                            s_star = s.idx

            w = w_min
            if w != float('inf'):
                w_stars.append(w)
                s_stars.append(s_star)
                s_star_names.append(self.strategies[s_star].name)

        return w_stars, s_star_names, s_stars

    def calculate_exp_incremental_nmbs(self, wtp_random_variate, n_samples, rnd):
        """ create summary statistics of incremental NMB of all strategies given the
        provided probability distribution of WTP value.
        :param wtp_random_variate: random variate generator for the probability distribution of wtp value
        :param n_samples: number of monte carlo samples
        :param rnd: the random number generator
        """

        for s in self.strategies[1:]:
            s.eIncNMB = utility_sample_stat(
                utility=self.utility,
                d_cost_samples=s.dCostObs,
                d_effect_samples=s.dEffectObs,
                wtp_random_variate=wtp_random_variate,
                n_samples=n_samples,
                rnd=rnd
            )

    def report_exp_incremental_nmb(self, interval='c', deci=0):

        report = []
        for s in self.strategies[1:]:

            mean_and_interval = s.eIncNMB.get_formatted_mean_and_interval(
                     interval_type=interval,
                     deci=deci,
                     form=','
                 )

            report.append([s.name, mean_and_interval])

        return report

    def plot_exp_incremental_nmb(self,
                                 title=None,
                                 y_label='Expected Net Monetary Benefit',
                                 y_range=None,
                                 figure_size=(5, 5),
                                 if_show_conf_interval=True,
                                 file_name=None):

        # initialize plot
        fig, ax = plt.subplots(figsize=figure_size)

        ax.set_title(title)
        ax.set_ylabel(y_label)

        for s in self.strategies[1:]:
            ax.scatter(x=s.idx, y=s.eIncNMB.get_mean(),
                       c=s.color, label=s.label)

            interval = s.eIncNMB.get_interval(interval_type='p')
            y = s.eIncNMB.get_mean()
            err = [[y-interval[0]], [interval[1]-y]]

            ax.errorbar(x=s.idx, y=y,
                        yerr=err,
                        c=s.color)

            if if_show_conf_interval:
                width = 0.1
                interval = s.eIncNMB.get_interval(interval_type='c')
                xy = [
                    [s.idx, interval[0]],
                    [s.idx - width, y],
                    [s.idx, interval[1]],
                    [s.idx + width, y]
                ]
                ax.add_patch(patches.Polygon(xy=xy, fill=False, edgecolor=s.color))

        ax.set_xticks([])
        ax.set_xlim([0.5, len(self.strategies)-0.5])
        ax.axhline(y=0, c='k', ls='--', lw=1)

        # format y-axis
        ax.set_ylim(y_range)
        vals_y = ax.get_yticks()
        ax.set_yticks(vals_y)
        ax.set_yticklabels(['{:,.{prec}f}'.format(x, prec=0) for x in vals_y])

        ax.legend()
        if file_name is None:
            fig.show()
        else:
            fig.savefig(file_name, bbox_inches='tight', dpi=300)


class BCHO(_EconEval):
    """ budget-constrained health optimization
    a class for selecting the alternative with the highest expected health outcomes
    subject to a budget constraint """

    def __init__(self, strategies, budget_range, if_paired, health_measure='u',
                 n_of_budget_values=200, epsilon=None):
        """
        :param strategies: the list of strategies (assumes that the first strategy represents the "base" strategy)
        :param budget_range: ([l, u]) range of budget values over which the analysis should be done
        :param if_paired: indicate whether the strategies are paired
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        :param n_of_budget_values: number of budget values to construct curves of optimal strategies
        :param epsilon: (float, 0 <= epsilon <= 1)
                        decision maker's tolerance in violating the budget constraint.
                        (i.e. epsilon in Prob{DCost_i > B} <= epsilon)
                        If set to None, this constraint will be considered: E[DCost_i] <= B.
        """
        _EconEval.__init__(self, strategies=strategies,
                           if_paired=if_paired,
                           health_measure=health_measure)

        # shift the strategies
        self._find_shifted_strategies()

        # list of cost values which the budget should be below
        # if epsilon = None, it takes the expected dCost of strategies,
        # otherwise, it takes the upper percentile of dCost of strategies.
        self.dCostUp = []
        # list of expected delta effect
        self.dEffect = []
        # list of expected effect curves (and EVPI if needed)
        self.curves = []
        # expected value of perfect information
        self.evpi = None

        # determine budget values
        self.budget_values = np.linspace(budget_range[0], budget_range[1],
                                         num=n_of_budget_values, endpoint=True)

        # set up curves
        for s in strategies:
            if epsilon is None:
                self.dCostUp.append(s.dCost.get_mean())
            else:
                self.dCostUp.append(s.dCost.get_percentile(q=(1 - epsilon) * 100))
            self.dEffect.append(s.dEffect.get_mean())
            self.curves.append(
                ExpHealthCurve(
                    label=s.name,
                    short_label=s.shortLabel,
                    color=s.color,
                    effect_stat=s.dEffect,
                    interval_type='c')
            )

        for b in self.budget_values:
            max_effect = -float('inf')
            max_s_i = None
            for s_i, s in enumerate(self.strategies):
                # if this strategy is feasible
                if self.dCostUp[s_i] <= b:
                    self.curves[s_i].update_feasibility(b=b)
                    if self.dEffect[s_i] > max_effect:
                        max_effect = self.dEffect[s_i]
                        max_s_i = s_i

            if max_s_i is None:
                self.curves[0].maxXs.append(b)
                self.curves[0].maxYs.append(None)
            else:
                self.curves[max_s_i].maxXs.append(b)
                self.curves[max_s_i].maxYs.append(max_effect)

        # convert lists to arrays
        for c in self.curves:
            c.convert_lists_to_arrays()

    def calculate_evpi_curve(self):
        """ calculates the expected value of perfect information (EVPI) curve """

        self.evpi = []
        n_of_sims = len(self.strategies[0].dCostObs)

        extra_budget = []
        for s in self.strategies:
            extra_budget.append(s.cost.get_percentile(97.5) - s.cost.get_mean())

        # for all budget value
        for b in self.budget_values:

            # find the best achievable expected effect under perfect information
            max_effects = []
            for i in range(n_of_sims):
                # find costs and effects of strategies for the ith monte carlo simulation run
                costs = [s.dCostObs[i] for s in self.strategies]
                effects = [s.dEffectObs[i] for s in self.strategies]

                # find the maximum effect
                max_e = float('-inf')
                for c, e, extra in zip(costs, effects, extra_budget):

                    # cost of this strategy doesn't satisfy the budget if
                    if c <= b + extra and e > max_e:
                        max_e = e
                max_effects.append(max_e)

            self.evpi.append(average(max_effects))

        # curve
        self.curves.append(EVPI(xs=self.budget_values, ys=self.evpi, label='EVPI', color='k'))

    def plot(self,
             title='Expected Increase in Effect',
             x_label='Budget',
             y_label='Expected Increase in Effect',
             show_evpi=False,
             y_range=None,
             y_axis_multiplier=1,
             delta_budget=None,
             transparency_lines=0.5,
             transparency_intervals=0.2,
             show_legend=True,
             figure_size=(5, 5),
             file_name='Budget.png'):

        # initialize plot
        fig, ax = plt.subplots(figsize=figure_size)

        if show_evpi:
            self.calculate_evpi_curve()

        # add plot to the ax
        self._add_curves_to_ax(ax=ax,
                               curves=self.curves,
                               x_values=self.budget_values,
                               title=title, x_label=x_label,
                               y_label=y_label, y_range=y_range, x_delta=delta_budget,
                               y_axis_multiplier=y_axis_multiplier,
                               transparency_lines=transparency_lines,
                               transparency_intervals=transparency_intervals,
                               show_legend=show_legend,
                               if_format_y_numbers=False)

        fig.show()
        if file_name is not None:
            fig.savefig(file_name, dpi=300)

    def add_plot_to_ax(self, ax, title=None,
                       delta_budget=None, x_label=None,
                       y_label=None, y_range=None, y_axis_multiplier=1, effect_decimals=None,
                       show_evpi=False, show_legend=True, show_frontier=True,
                       ):
        """
        :param ax: axis
        :param title: (string) title of the figure
        :param delta_budget: (float) the distance between ticks on the x-axis
        :param x_label: (string) x-axis label
        :param y_label: (string) y-axis label
        :param y_range: (tuple) y-axis range
        :param y_axis_multiplier: (float) to multiply the y-axis values by
        :param effect_decimals: (int) to round the values of y-axis (effect)
        :param show_evpi: (bool) to show the expected value of perfect information (EVPI) curve
        :param show_legend: (bool) to show legends
        :param show_frontier: (bool) to show the frontier (curves with maximum effect or NMB)
        :return:
        """

        if show_evpi:
            self.calculate_evpi_curve()

        add_curves_to_ax(
            ax=ax, curves=self.curves, title=title,
            x_values=self.budget_values,
            x_delta=delta_budget, x_label=x_label,
            y_label=y_label, y_axis_decimal=effect_decimals, y_range=y_range, y_axis_multiplier=y_axis_multiplier,
            transparency_lines=1, transparency_intervals=NMB_INTERVAL_TRANSPARENCY,
            show_legend=show_legend,
            show_frontier=show_frontier,
            curve_line_width=NMB_LINE_WIDTH, frontier_line_width=NMB_FRONTIER_LINE_WIDTH,
            if_format_y_numbers=True if effect_decimals is not None else False,
            legend_font_size=LEGEND_FONT_SIZE,
            frontier_label_shift_x=FRONTIER_LABEL_SHIFT_X,
            frontier_label_shift_y=FRONTIER_LABEL_SHIFT_Y
        )


class _ComparativeEconMeasure:
    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) effect data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) effect data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        if health_measure not in ['u', 'd']:
            raise ValueError("health_measure can be either 'u' (for utility) or 'd' (for disutility).")

        self.name = name
        # if QALY or DALY is being used
        self._effect_multiplier = 1 if health_measure == 'u' else -1

        # convert input data to numpy.array if needed
        self._costsNew = assert_np_list(costs_new, "cost_new should be list or np.array.")
        self._effectsNew = assert_np_list(effects_new, "effects_new should be list or np.array.")
        self._costsBase = assert_np_list(costs_base, "costs_base should be list or np.array.")
        self._effectsBase = assert_np_list(effects_base, "effects_base should be list or np.array.")

        # calculate the difference in average cost
        self._delta_ave_cost = np.average(self._costsNew) - np.average(self._costsBase)
        # change in effect: DALY averted or QALY gained
        self._delta_ave_effect = (np.average(self._effectsNew) - np.average(self._effectsBase)) \
                                 * self._effect_multiplier

    def get_ave_d_cost(self):
        """
        :return: average incremental cost
        """
        return self._delta_ave_cost

    def get_ave_d_effect(self):
        """
        :return: average incremental effect
        """
        return self._delta_ave_effect


class _ICER(_ComparativeEconMeasure):
    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) effect data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) effect data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        self._isDefined = True  # if ICER cannot be computed, this will change to False

        # initialize the base class
        _ComparativeEconMeasure.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

        # calculate ICER
        if not (self._delta_ave_effect > 0 and self._delta_ave_cost >= 0):
            warnings.warn(self.name + ': Mean incremental effect should be > 0 '
                                      'and mean incremental cost should be >= 0. '
                                      'ICER is not computable.')
            self._isDefined = False
            self._ICER = math.nan
        else:
            # $ per DALY averted or $ per QALY gained
            self._ICER = self._delta_ave_cost / self._delta_ave_effect


    def get_ICER(self):
        """ return ICER """
        return self._ICER

    def get_CI(self, alpha=0.05, method='bootstrap', num_bootstrap_samples=1000, rng=None,
               prior_range=None, num_wtp_thresholds=1000):
        """
        :param alpha: significance level, a value from [0, 1]
        :param method: (string) 'bootstrap' or 'Bayesian'
        :param num_bootstrap_samples: number of bootstrap samples when 'bootstrap' method is selected
        :param rng: random number generator to generate empirical bootstrap samples
        :param num_wtp_thresholds: (int) number of willingness-to-pay thresholds to evaluate posterior
            when 'Bayesian' approach is selected
        :param prior_range: (tuple) in form of (l, u) for the prior range of willingness-to-pay
            threshold that makes NMB zero (if prior is not provided [0, 4 * ICER] will be used.
        :return: confidence interval in the format of list [l, u]
        """
        # abstract method to be overridden in derived classes to process an event
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_PI(self, alpha=0.05):
        """
        :param alpha: significance level, a value from [0, 1]
        :return: percentile interval in the format of list [l, u]
        """
        # abstract method to be overridden in derived classes to process an event
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_formatted_mean_and_interval(self, interval_type='c',
                                        alpha=0.05, method='bootstrap', num_bootstrap_samples=1000, rng=None,
                                        prior_range=None, num_wtp_thresholds=1000,
                                        deci=0, sig_digits=4, form=None,
                                        multiplier=1):
        """
        :param interval_type: (string) 'n' for no interval
                                       'c' or 'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param alpha: significance level, a value from [0, 1]
        :param method: (string) method to calculate confidence interval ('bootstrap' or 'Bayesian')
        :param num_bootstrap_samples: number of bootstrap samples when 'bootstrap' method is selected
        :param rng: random number generator to generate empirical bootstrap samples
        :param num_wtp_thresholds: (int) number of willingness-to-pay thresholds to evaluate posterior
            when 'Bayesian' approach is selected
        :param prior_range: (tuple) in form of (l, u) for the prior range of willingness-to-pay
            threshold that makes NMB zero (if prior is not provided [0, 4 * ICER] will be used
        :param deci: digits to round the numbers to
        :param sig_digits: number of significant digits
        :param form: ',' to format as number, '%' to format as percentage, and '$' to format as currency
        :param multiplier: to multiply the estimate and the interval by the provided value
        :param num_bootstrap_samples: number of bootstrap samples to calculate confidence interval
        :return: (string) estimate of ICER and interval formatted as specified
        """

        estimate = self.get_ICER() * multiplier

        if interval_type == 'c' or interval_type == 'cb':
            interval = self.get_CI(alpha=alpha, method=method, num_bootstrap_samples=num_bootstrap_samples, rng=rng,
                                   prior_range=prior_range, num_wtp_thresholds=num_wtp_thresholds)
        elif interval_type == 'p':
            interval = self.get_PI(alpha=alpha)
        else:
            interval = None

        adj_interval = [v * multiplier for v in interval] if interval is not None else None

        return F.format_estimate_interval(estimate=estimate,
                                          interval=adj_interval,
                                          deci=deci,
                                          sig_digits=sig_digits,
                                          format=form)


class ICER_Paired(_ICER):

    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) health data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the baseline
        :param effects_base: (list or numpy.array) health data for the baseline
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        # all cost and effects should have the same length
        if not (len(costs_new) == len(effects_new) == len(costs_base) == len(effects_base)):
            raise ValueError('Paired ICER assume the same number of observations for all cost and health outcomes.')

        # initialize the base class
        _ICER.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

        # incremental observations
        self._deltaCosts = self._costsNew - self._costsBase
        self._deltaEffects = (self._effectsNew - self._effectsBase) * self._effect_multiplier

    def get_CI(self, alpha=0.05, method='bootstrap', num_bootstrap_samples=1000, rng=None,
               prior_range=None, num_wtp_thresholds=1000):
        """
        :param alpha: significance level, a value from [0, 1]
         :param method: (string) 'bootstrap' or 'Bayesian'
        :param num_bootstrap_samples: number of bootstrap samples when 'bootstrap' method is selected
        :param rng: random number generator to generate empirical bootstrap samples
        :param num_wtp_thresholds: (int) number of willingness-to-pay thresholds to evaluate posterior
            when 'Bayesian' approach is selected
        :param prior_range: (tuple) in form of (l, u) for the prior range of willingness-to-pay
            threshold that makes NMB zero (if prior is not provided [0, 4 * ICER] will be used.
        :return: confidence interval in the format of list [l, u]
        """

        n_obs = len(self._deltaCosts)

        # create a new random number generator if one is not provided.
        if rng is None:
            rng = RandomState(seed=1)

        # check if the Bayesian approach is selected
        if method == 'Bayesian':

            if prior_range is None:
                prior_range = [0, 4*self._ICER]

            # lambda0 s
            lambda_0s = np.linspace(start=prior_range[0],
                                    stop=prior_range[1],
                                    num=num_wtp_thresholds)
            lnl_weights = []
            mean_d_effect = self._delta_ave_effect
            st_err_d_effect = sem(self._deltaEffects)
            mean_d_cost = self._delta_ave_cost
            st_err_d_cost = sem(self._deltaCosts)

            # lnl of observing NMB = 0 given the sampled lambda_0s
            for lambda_0 in lambda_0s:
                lnl_weight = 0
                lnl_weight += stat.norm.logpdf(
                    x=0,
                    loc=lambda_0 * mean_d_effect - mean_d_cost,
                    scale=lambda_0 * st_err_d_effect + st_err_d_cost)

                lnl_weights.append(lnl_weight)

            # convert likelihoods to probabilities
            probs = convert_lnl_to_prob(lnl_weights)

            # l = get_percentile_of_empirical_dist(xs=lambda_0s, probs=probs, q=alpha / 2)
            # u = get_percentile_of_empirical_dist(xs=lambda_0s, probs=probs, q=1 - alpha / 2)
            # print([l, u])

            # resamples lambda_0s based on the probabilities
            sampled_lambda_0s = np.random.choice(
                a=lambda_0s,
                size=num_wtp_thresholds,
                replace=True,
                p=probs)

            # report CI
            sum_stat = SummaryStat(data=sampled_lambda_0s)
            return sum_stat.get_interval(interval_type='p', alpha=alpha)

        elif method == 'bootstrap':
            # bootstrap algorithm
            icer_bootstrap_means = np.zeros(num_bootstrap_samples)
            for i in range(num_bootstrap_samples):
                # because cost and health observations are paired,
                # we sample delta cost and delta health together
                indices = rng.choice(a=range(n_obs),
                                     size=n_obs,
                                     replace=True)
                sampled_delta_costs = self._deltaCosts[indices]
                sampled_delta_effects = self._deltaEffects[indices]

                ave_delta_cost = np.average(sampled_delta_costs)
                ave_delta_effect = np.average(sampled_delta_effects)

                # assert all the means should not be 0
                if ave_delta_effect <= 0:
                    warnings.warn(
                        self.name + ': Mean incremental health is 0 or less for one bootstrap sample, '
                                    'ICER is not computable')
                    self._isDefined = False
                    return [math.nan, math.nan]

                icer_bootstrap_means[i] = ave_delta_cost / ave_delta_effect - self._ICER

            # return the bootstrap interval
            return self._ICER - np.percentile(icer_bootstrap_means, [100 * (1 - alpha / 2.0), 100 * alpha / 2.0])

        else:
            raise ValueError('Invalid method. Method should be either bootstrap or Bayesian.')

    def get_PI(self, alpha=0.05):
        """
        :param alpha: significance level, a value from [0, 1]
        :return: prediction interval in the format of list [l, u]
        """

        # calculate ICERs
        if min(self._deltaEffects) <= 0:
            warnings.warn("\nFor '{0},' the prediction interval of ICERs is not computable because at least one "
                          "incremental effect is negative or zero.".format(self.name))
            return [math.nan, math.nan]
        else:
            icers = np.divide(self._deltaCosts, self._deltaEffects)

        return np.percentile(icers, [100 * alpha / 2.0, 100 * (1 - alpha / 2.0)])


class ICER_Indp(_ICER):

    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) health data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) health data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        # all cost and effects should have the same length for each alternative
        if not (len(costs_new) == len(effects_new) and len(costs_base) == len(effects_base)):
            raise ValueError(
                'ICER assume the same number of observations for the cost and health outcome of each alternative.')

        # initialize the base class
        _ICER.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

    def get_CI(self, alpha=0.05, method='bootstrap', num_bootstrap_samples=1000, rng=None,
               prior_range=None, num_wtp_thresholds=1000):
        """
        :param alpha: significance level, a value from [0, 1]
        :param method: (string) 'bootstrap' or 'Bayesian'
        :param num_bootstrap_samples: number of bootstrap samples when 'bootstrap' method is selected
        :param rng: random number generator to generate empirical bootstrap samples
        :param num_wtp_thresholds: (int) number of willingness-to-pay thresholds to evaluate posterior
            when 'Bayesian' approach is selected
        :param prior_range: (tuple) in form of (l, u) for the prior range of willingness-to-pay
            threshold that makes NMB zero (if prior is not provided [0, 4 * ICER] will be used.
        :return: confidence interval in the format of list [l, u]
        """

        if not self._isDefined:
            return [math.nan, math.nan]

        if method == 'Bayesian':
            raise ValueError('The Bayesian approach is not yet implemented. Use bootsrap instead.')

        # create a new random number generator if one is not provided.
        if rng is None:
            rng = RandomState(seed=1)

        # vector to store bootstrap ICERs
        icer_bootstrap_means = np.zeros(num_bootstrap_samples)

        n_obs_new = len(self._costsNew)
        n_obs_base = len(self._costsBase)

        # get bootstrap samples
        for i in range(num_bootstrap_samples):
            # for the new alternative
            indices_new = rng.choice(a=range(n_obs_new), size=n_obs_new, replace=True)
            costs_new = self._costsNew[indices_new]
            effects_new = self._effectsNew[indices_new]

            # for the base alternative
            indices_base = np.random.choice(a=range(n_obs_base), size=n_obs_base, replace=True)
            costs_base = self._costsBase[indices_base]
            effects_base = self._effectsBase[indices_base]

            # calculate this bootstrap ICER
            mean_costs_new = np.mean(costs_new)
            mean_costs_base = np.mean(costs_base)
            mean_effects_new = np.mean(effects_new)
            mean_effects_base = np.mean(effects_base)

            # calculate this bootstrap ICER
            if (mean_effects_new - mean_effects_base) * self._effect_multiplier <= 0:
                self._isDefined = False
                warnings.warn('\nFor "{}, the confidence interval of ICER is not computable."'
                              '\nThis is because at least one of bootstrap mean incremental effect '
                              'is negative.'
                              '\nIncreasing the number of cost and effect observations '
                              'might resolve the issue.'.format(self.name))
                return [math.nan, math.nan]

            else:
                icer_bootstrap_means[i] = \
                    (mean_costs_new - mean_costs_base) / (mean_effects_new - mean_effects_base) \
                    * self._effect_multiplier

        if self._isDefined:
            return np.percentile(icer_bootstrap_means, [100 * alpha / 2.0, 100 * (1 - alpha / 2.0)])
        else:
            return [math.nan, math.nan]

    def get_PI(self, alpha=0.05, num_bootstrap_samples=1000, rng=None):
        """
        :param alpha: significance level, a value from [0, 1]
        :param num_bootstrap_samples: number of bootstrap samples
        :param rng: random number generator
        :return: prediction interval in the format of list [l, u]
        """

        if self._ICER is math.nan:
            return [math.nan, math.nan]

        # create a new random number generator if one is not provided.
        if rng is None:
            rng = RandomState(seed=1)

        if num_bootstrap_samples == 0:
            num_bootstrap_samples = max(len(self._costsNew), len(self._costsBase))

        # calculate element-wise ratio as sample of ICER
        indices_new = rng.choice(a=range(num_bootstrap_samples), size=num_bootstrap_samples, replace=True)
        costs_new = self._costsNew[indices_new]
        effects_new = self._effectsNew[indices_new]

        indices_base = rng.choice(a=range(num_bootstrap_samples), size=num_bootstrap_samples, replace=True)
        costs_base = self._costsBase[indices_base]
        effects_base = self._effectsBase[indices_base]

        if min((effects_new - effects_base) * self._effect_multiplier) <= 0:
            self._isDefined = False
            warnings.warn('\nFor "{}, the prediction interval of ICER is not computable."'
                          '\nThis is because at least one of bootstrap mean incremental effect '
                          'is negative'.format(self.name))
            return [math.nan, math.nan]
        else:
            sample_icers = np.divide(
                (costs_new - costs_base),
                (effects_new - effects_base) * self._effect_multiplier)

        if self._isDefined:
            return np.percentile(sample_icers, [100 * alpha / 2.0, 100 * (1 - alpha / 2.0)])
        else:
            return [math.nan, math.nan]


class _INMB(_ComparativeEconMeasure):
    # incremental net monetary benefit
    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) effect data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) effect data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """
        # initialize the base class
        _ComparativeEconMeasure.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

    def get_INMB(self, wtp):
        """
        :param wtp: willingness-to-pay ($ for QALY gained or $ for DALY averted)
        :returns: the incremental net monetary benefit at the provided willingness-to-pay value
        """
        return wtp * self._delta_ave_effect - self._delta_ave_cost

    def get_CI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: confidence interval in the format of list [l, u]
        """
        # abstract method to be overridden in derived classes
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_PI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: percentile interval in the format of list [l, u]
        """
        # abstract method to be overridden in derived classes
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_switch_wtp(self):

        try:
            wtp = self.get_ave_d_cost() / self.get_ave_d_effect()
        except ValueError:
            wtp = math.nan

        return wtp

    def get_switch_wtp_and_interval(self, wtp_range, interval_type='n'):

        wtp = self.get_switch_wtp()

        if interval_type == 'n':
            return wtp, None
        elif interval_type == 'c':
            interval_at_min_wtp = self.get_CI(wtp=wtp_range[0])
            interval_at_max_wtp = self.get_CI(wtp=wtp_range[1])
        elif interval_type == 'p':
            interval_at_min_wtp = self.get_PI(wtp=wtp_range[0])
            interval_at_max_wtp = self.get_PI(wtp=wtp_range[1])
        else:
            raise ValueError('Invalid value for interval_type.')

        line_lower_err = Line(x1=wtp_range[0],
                              x2=wtp_range[1],
                              y1=interval_at_min_wtp[0],
                              y2=interval_at_max_wtp[0])
        line_upper_err = Line(x1=wtp_range[0],
                              x2=wtp_range[1],
                              y1=interval_at_min_wtp[1],
                              y2=interval_at_max_wtp[1])

        if self.get_ave_d_effect() >= 0:
            interval = [line_upper_err.get_intercept_with_x_axis(),
                        line_lower_err.get_intercept_with_x_axis()]
        else:
            interval = [line_lower_err.get_intercept_with_x_axis(),
                        line_upper_err.get_intercept_with_x_axis()]

        return wtp, interval


class INMB_Paired(_INMB):

    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) health data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) health data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        # all cost and effects should have the same length
        if not (len(costs_new) == len(effects_new) == len(costs_base) == len(effects_base)):
            raise ValueError(
                'Paired incremental NMB assumes the same number of observations for all cost and health outcomes.')

        _INMB.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

        # incremental observations
        self._deltaCost = self._costsNew - self._costsBase
        self._deltaHealth = (self._effectsNew - self._effectsBase) * self._effect_multiplier

        self._n = len(costs_new)
        self._statDeltaCost = Stat.SummaryStat(name=self.name, data=self._deltaCost)
        self._statDeltaHealth = Stat.SummaryStat(name=self.name, data=self._deltaHealth)

    def get_CI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: confidence interval in the format of list [l, u]
        """
        mean = self.get_INMB(wtp=wtp)

        t = math.nan
        if self._n > 1:
            t = stat.t.ppf(1 - alpha / 2, self._n - 1)

        st_dev = math.sqrt(wtp ** 2 * self._statDeltaHealth.get_var() + self._statDeltaCost.get_var())
        st_err = st_dev/math.sqrt(self._n)

        l = mean - t * st_err
        u = mean + t * st_err
        return [l, u]

    def get_PI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: percentile interval in the format of list [l, u]
        """
        return Stat.SummaryStat(name=self.name,
                                data=wtp * self._deltaHealth - self._deltaCost).get_PI(alpha)


class INMB_Indp(_INMB):

    def __init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure='u'):
        """
        :param costs_new: (list or numpy.array) cost data for the new strategy
        :param effects_new: (list or numpy.array) health data for the new strategy
        :param costs_base: (list or numpy.array) cost data for the base line
        :param effects_base: (list or numpy.array) health data for the base line
        :param health_measure: (string) choose 'u' if higher "effect" implies better health
        (e.g. when QALY is used) and set to 'd' if higher "effect" implies worse health
        (e.g. when DALYS is used)
        """

        # all costs and effects should have the same length for each strategy
        if not (len(costs_new) == len(effects_new) and len(costs_base) == len(effects_base)):
            raise ValueError(
                'Independent incremental NMB assumes that for each strategy there are '
                'the same number of observations for cost and health outcomes.')

        _INMB.__init__(self, name, costs_new, effects_new, costs_base, effects_base, health_measure)

    def get_CI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: confidence interval in the format of list [l, u]
        """
        # NMB observations of two alternatives
        stat_new = wtp * self._effectsNew * self._effect_multiplier - self._costsNew
        stat_base = wtp * self._effectsBase * self._effect_multiplier - self._costsBase

        # to get CI for stat_new - stat_base
        diff_stat = Stat.DifferenceStatIndp(name=self.name, x=stat_new, y_ref=stat_base)
        return diff_stat.get_t_CI(alpha)

    def get_PI(self, wtp, alpha=0.05):
        """
        :param wtp: willingness-to-pay value ($ for QALY gained or $ for DALY averted)
        :param alpha: significance level, a value from [0, 1]
        :return: percentile interval in the format of list [l, u]
        """
        # NMB observations of two alternatives
        stat_new = wtp * self._effectsNew * self._effect_multiplier - self._costsNew
        stat_base = wtp * self._effectsBase * self._effect_multiplier - self._costsBase

        # to get PI for stat_new - stat_base
        diff_stat = Stat.DifferenceStatIndp(name=self.name, x=stat_new, y_ref=stat_base)
        return diff_stat.get_PI(alpha)




import matplotlib.pyplot as plt
import numpy as np

import deampy.regression as Reg

# generate x values (observations)
x = np.random.randn(20) # x = np.linspace(100000, 150000, 20)
# generate y values (assuming y = x^2 -2 + error)
y = pow(x, 3) - 4 * pow(x, 2) + x - 2 + 10*np.random.randn(len(x))

# create the scatter plot of (x, y) points
fig, ax = plt.subplots(figsize=(8, 4))
ax.scatter(x, y, alpha=0.5, color='orchid')
fig.suptitle('Example Scatter Plot')
fig.tight_layout(pad=2)
ax.grid(True)

# create the regression model
single_var_poly_reg = Reg.SingleVarPolyRegWithInference(x, y, degree=3)

# print the coefficients of the fitted model
print('Estimated coefficients:', single_var_poly_reg.get_coeffs())

print(single_var_poly_reg.fitted.summary())
print(single_var_poly_reg.fitted.pvalues)

# print derivative
print('Derivative at x=1:', single_var_poly_reg.get_derivative(x=1))

# print zeros
print('x for which f(x) = 0:', single_var_poly_reg.get_zero())

# make prediction over the range [x_min, x_max]
x_pred = np.linspace(x.min(), x.max(), 50)
y_pred = single_var_poly_reg.get_predicted_y(x_pred)

# plot the predicted values
ax.plot(x_pred, y_pred, '-', color='darkorchid', linewidth=2)

# add the confidence region of predictions
lower, upper = single_var_poly_reg.get_predicted_y_CI(x_pred)
ax.fill_between(x_pred, lower, upper, color='#888888', alpha=0.4)

# add the prediction intervals of predictions
lower, upper = single_var_poly_reg.get_predicted_y_PI(x_pred)
ax.fill_between(x_pred, lower, upper, color='#888888', alpha=0.1)

plt.show()



import deampy.format_functions as Format

print("\nFormatting estimates:")
print(Format.format_number(12345.678, deci=1))
print(Format.format_number(12345.678, deci=1, format=','))
print(Format.format_number(12345.678, deci=1, format='$'))
print(Format.format_number(0.12345, deci=1, format='%'))

print("\nFormatting intervals:")
print(Format.format_interval([12345.678, 98765.432], deci=1))
print(Format.format_interval([12345.678, 98765.432], deci=1, format=','))
print(Format.format_interval([12345.678, 98765.432], deci=1, format='$'))
print(Format.format_interval([0.12345, 0.98765], deci=1, format='%'))

print("\nFormatting estimates and intervals:")
print(Format.format_estimate_interval(
    estimate=5555.55, interval=[12345.678, 98765.432], deci=1))
print(Format.format_estimate_interval(
    estimate=5555.55, interval=[12345.678, 98765.432], deci=1, format=','))
print(Format.format_estimate_interval(
    estimate=5555.55, interval=[12345.678, 98765.432], deci=1, format='$'))
print(Format.format_estimate_interval(
    estimate=0.5555, interval=[0.12345, 0.98765], deci=1, format='%'))



import numpy as np

import deampy.markov as Cls

myGillespie = Cls.Gillespie(transition_rate_matrix=
                            [[None, 0.1],
                             [0, None]])

t = 0
i = 0
rng = np.random.RandomState()
while t is not None:
    print('Current state:', i)
    t, i = myGillespie.get_next_state(current_state_index=i, rng=rng)
    print('Time to next transition:', t)




import numpy as np

from deampy import econ_eval as EV

np.random.seed(seed=1)

S0 = EV.Strategy(name='Base',
                 cost_obs=np.random.normal(loc=100, scale=10, size=100),
                 effect_obs=np.random.normal(loc=1, scale=.2, size=100),
                 color='red')
S1 = EV.Strategy(name='A1',
                 cost_obs=np.random.normal(loc=800, scale=50, size=100),
                 effect_obs=np.random.normal(loc=0.5, scale=0.1, size=100),
                 color='blue')
S2 = EV.Strategy(name='A2',
                 cost_obs=np.random.normal(loc=2000, scale=200, size=100),
                 effect_obs=np.random.normal(loc=10, scale=1, size=100),
                 color='green')
S3 = EV.Strategy(name='A3',
                 cost_obs=np.random.normal(loc=500, scale=50, size=100),
                 effect_obs=np.random.normal(loc=7, scale=1, size=100),
                 color='orange')
S4 = EV.Strategy(name='A4',
                 cost_obs=np.random.normal(loc=-100, scale=10, size=100),
                 effect_obs=np.random.normal(loc=2, scale=0.1, size=100),
                 color='purple')

cea = EV.CEA(strategies=[S0, S1, S2, S3, S4], if_paired=False, health_measure='u')

print('On frontier')
for s in cea.get_strategies_on_frontier():
    print(s.name)

print('Not on frontier')
for s in cea.get_strategies_not_on_frontier():
    print(s.name)

cea.plot_CE_plane(add_clouds=True)
cea.build_CE_table(cost_digits=1, interval_type='c')

#cea.print_pairwise_cea(interval_type='c')
cea.plot_pairwise_ceas(
    figure_size=(7, 7),
    font_size=7,
    effect_label='DALY Averted (Thousands)',
    cost_label='Additional Cost (Thousand Dollars)',
    center_s=40,
    cloud_s=10,
    transparency=0.1,
    x_range=[-20, 20],
    y_range=[-2500, 2500],
)



from deampy.data_structure import Pyramid

from deampy.plots import pop_pyramids as P

"""
a pyramid to store population size by age and sex:
    age,   sex,      population
    0,     0,        10,
    0,     1,        20,
    5,     0,        30,
    5,     1,        40,
    10,    0,        50,
    10,    1,        60
"""

pyramid = Pyramid(list_x_min=[0, 0],
                  list_x_max=[10, 1],
                  list_x_delta=[5, 'int'])

# adding a person of age 1.2 and sex = 0
pyramid.record_increment(x_values=[1.2, 0], increment=1)
# adding 10 persons of age 5 and sex = 1
pyramid.record_increment(x_values=[5, 1], increment=10)
# removing 2 persons of age 5.2 and sex = 1
pyramid.record_increment(x_values=[5.2, 1], increment=-2)
# adding 1 persons of age 15 and sex = 0
pyramid.record_increment(x_values=[15, 0], increment=1)

print('Size of group with age 0-5 and sex 0:', pyramid.get_current_value(x_values=[0, 0]))
print('Size of group with age 5-10 and sex 1:', pyramid.get_current_value(x_values=[5, 1]))
# get the total population size
print('Population size:', pyramid.get_sum())
# get the size of each group
print('Population size by age, sex:', pyramid.get_values())
# get the percentage of population in each group
print('Population distribution by age, sex', pyramid.get_percentages())
# plot the pyramid
P.plot_pyramids(observed_data=pyramid.get_percentages(),
                simulated_data=None,
                x_lim=100,
                title='Testing Pyramid')


# updating a new pyramid based on the one above
newPyramid = Pyramid(list_x_min=[0, 0],
                     list_x_max=[10, 1],
                     list_x_delta=[5, 'int'])
newPyramid.record_values_from_another_pyramid(another_pyramid=pyramid)

print('\nTesting the new pyramid:')
# get the total population size
print('Population size:', newPyramid.get_sum())
# get the size of each group
print('Population size by age, sex:', newPyramid.get_values())
# get the percentage of population in each group
print('Population distribution by age, sex', newPyramid.get_percentages())



import matplotlib.pyplot as plt
import numpy as np

import deampy.plots.plot_support as Fig

data = [1, None, 3, 4, 5, 6, 7, None, 9]
print(Fig.get_moving_average(data=data, window=3))


data = np.random.randn(100)

fig = plt.figure(figsize=(4,2))
ax = fig.add_subplot(111)
ax.plot(range(100), data, alpha=0.5)
ax.plot(range(100), Fig.get_moving_average(data, window=10), alpha=1)

plt.show()



import deampy.random_variats as RVGs
from deampy.data_structure import DataFrameOfObjects


class MortalityModel:
    """
    example:
        age,   sex,      mortality rate
        0,     0,        0.1,
        0,     1,        0.11,
        5,     0,        0.2,
        5,     1,        0.21,
        10,    0,        0.3
        10,    1,        0.31

    This class assumes that the first column contains age groups and the last column contains mortality rates
    """

    def __init__(self, mortality_rates, group_mins, group_maxs, group_delta, age_min, age_delta):
        """
        :param mortality_rates: (list of list) the table above
        :param group_mins: list of minimum value of x (in example above: [0, 0])
        :param group_maxs: list of maximum value of x (in example above: [10, 1])
        :param group_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        :param age_min:
        :param age_delta:
        """

        self.df = DataFrameOfObjects(list_x_min=group_mins,
                                     list_x_max=group_maxs,
                                     list_x_delta=group_delta)

        self.ageMin = age_min

        for df_row in self.df.get_rows():
            rates = []
            for row in mortality_rates:
                if df_row[:-1] == row[1:-1]:
                    rates.append(row[-1])
            self.df.set_obj(x_value=df_row[0:-1],
                            obj=RVGs.NonHomogeneousExponential(rates=rates, delta_t=age_delta))

    def sample_time_to_death(self, group, age, rng):

        if age < self.ageMin:
            raise ValueError('Current age cannot be smaller than the minimum age.')

        return self.df.get_obj(group).sample(rng=rng, arg=age)



from deampy import in_out_functions as io

# ---------------
# first run the TestCSVWritter.py to produce the csv file
# ---------------

# test reading by rows
rows = io.read_csv_rows('CSVFolder/myCSV.csv',
                        if_ignore_first_row=True,
                        if_convert_float=True)
print('Testing reading by rows:')
for row in rows:
    print(sum(row[1:]))

# test reading by columns
cols = io.read_csv_cols('CSVFolder/myCSV.csv',
                        n_cols=4,
                        if_ignore_first_row=True,
                        if_convert_float=True)
print('Testing reading by columns:')
for j in range(1, 4):
    print(sum(cols[j]))

# rest reading by columns into a dictionary
dict_cols = io.read_csv_cols_to_dictionary(
    'CSVFolder/myCSV.csv',
    if_convert_float=True)

print('Testing reading by columns into a dictionary:')
print(dict_cols)


from numpy.random import rand

from deampy.sensitivity_analysis import SensitivityAnalysis

# prepare data
param1 = rand(100) * 20
param2 = rand(100) * -1
output = param1 + 2*param2 + rand(100) * 10

dic_par_values = {'Par1': param1, 'Par2': param2}

sa = SensitivityAnalysis(dic_parameter_values=dic_par_values, output_values=output)

sa.print_corr(corr='r')
sa.print_corr(corr='rho')
sa.print_corr(corr='p')
sa.print_corr(corr='pr')
sa.export_to_csv(corrs=['r', 'rho', 'p', 'pr'], decimal=4)



import numpy as np

import deampy.statistics as S
from deampy.random_variats import Beta


class Model:

    def __init__(self, decision_rule=None, cost_sigma=0.5, action_cost=1):

        self.costSigma = cost_sigma
        self.actionCost = action_cost

        self.decisionRule = decision_rule
        self.seqCosts = []

    def simulate(self, itr):

        self._reset()
        rng = np.random.RandomState(seed=itr)

        # find the initial state
        state = rng.random_sample()

        for t in range(3):

            # make a decision
            action = self.decisionRule.get_decision(feature_values=[t, state])

            # store the reward
            cost = self._cost(state=state, action=action, rng=rng)
            self.seqCosts.append(cost)

            # next state
            if action[0] == 0:
                result = Beta.fit_mm(mean=state, st_dev=0.01)
                if result['a'] <= 0 or result['b'] <= 0:
                    state = state
                else:
                    state = rng.beta(a=result['a'], b=result['b'])
            elif action[0] == 1:
                if state < 0.5:
                    state += (0.5 - state) * rng.random_sample()
                else:
                    state -= (state - 0.5) * rng.random_sample()
            else:
                raise ValueError

    def _cost(self, state, action, rng):

        state_cost = 1000*pow(state - 0.5, 2) + rng.normal(0, self.costSigma)
        action_cost = self.actionCost if action[0] == 1 else 0

        return state_cost + action_cost

    def set_approx_decision_maker(self, approx_decision_maker):

        self.decisionRule = Dynamic(approx_decision_maker=approx_decision_maker)

    def get_seq_of_costs(self):
        return self.seqCosts

    def _reset(self):
        self.seqCosts.clear()


class MultiModel:

    def __init__(self, decision_rule=None, cost_sigma=0.5, action_cost=1):

        self.decisionRule = decision_rule
        self.costSigma = cost_sigma
        self.actionCost = action_cost

        self.cumCosts = []
        self.statCost = None

    def simulate(self, n):

        for i in range(n):
            model = Model(decision_rule=self.decisionRule,
                          cost_sigma=self.costSigma,
                          action_cost=self.actionCost)
            model.simulate(itr=i)
            self.cumCosts.append(sum(model.seqCosts))

        self.statCost = S.SummaryStat(data=self.cumCosts)


class _DecisionRule:
    def __init__(self):
        pass

    def get_decision(self, feature_values):
        raise NotImplementedError


class AlwaysOn(_DecisionRule):
    def __init__(self):
        _DecisionRule.__init__(self)

    def get_decision(self, feature_values):
        return [1]


class AlwaysOff(_DecisionRule):
    def __init__(self):
        _DecisionRule.__init__(self)

    def get_decision(self, feature_values):
        return [0]


class Myopic(_DecisionRule):
    def __init__(self):
        _DecisionRule.__init__(self)

    def get_decision(self, feature_values):
        return [0] if 0.25 <= feature_values[1] <= 0.75 else [1]


class Dynamic(_DecisionRule):

    def __init__(self, approx_decision_maker):
        _DecisionRule.__init__(self)

        self.approxDecisionMaker = approx_decision_maker

    def get_decision(self, feature_values):

        return self.approxDecisionMaker.make_a_decision(continuous_feature_values=feature_values)




import deampy.regression as Reg

x = [1296, 680, 529, 471, 444, 426, 412, 407, 401, 238]
y = [1, 18, 29, 36, 42, 48, 57, 60, 67, 207]

reg = Reg.PowerRegression(x, y, if_c0_zero=True)
reg_with_p0 = Reg.PowerRegression(x, y, if_c0_zero=True, p0=(1, 0))
reg.plot_fit()
reg_with_p0.plot_fit()
print(reg.get_coeffs())
print('\ncoefficients of model without initial value:', reg.get_coeffs())
print('coefficients of model with initial value', reg_with_p0.get_coeffs())



from deampy import econ_eval as EV


S0 = EV.Strategy(name='Base', cost_obs=[100], effect_obs=[1])
S1 = EV.Strategy(name='A1', cost_obs=[800], effect_obs=[0.5])
S2 = EV.Strategy(name='A2', cost_obs=[2000], effect_obs=[10])
S3 = EV.Strategy(name='A3', cost_obs=[500], effect_obs=[7])
S4 = EV.Strategy(name='A4', cost_obs=[-100], effect_obs=[2])

cea = EV.CEA(strategies=[S0, S1, S2, S3, S4], if_paired=False, health_measure='u')

print('On frontier')
for s in cea.get_strategies_on_frontier():
    print(s.name)


print('Not on frontier')
for s in cea.get_strategies_not_on_frontier():
    print(s.name)


cea.plot_CE_plane(add_clouds=False)
cea.build_CE_table(cost_digits=0, interval_type='n')



import numpy as np

import deampy.econ_eval as EconEval

np.random.seed(573)

cost_base = np.random.normal(loc=10000, scale=100, size=1000)
effect_base = np.random.normal(loc=2, scale=.1, size=1000)
cost_intervention = np.random.normal(loc=20000, scale=200, size=1000)
effect_intervention = np.random.normal(loc=1, scale=.2, size=1000)

print('')

# ICER calculation assuming paired observations
ICER_paired = EconEval.ICER_Paired('Testing paired ICER',
                                   cost_intervention, effect_intervention, cost_base, effect_base,
                                   health_measure='d')
print('Paired ICER:\n\tICER: {} \n\tCI (boostrap): {} \n\tCI (Bayesian): {} \n\tPI: '.format(
      ICER_paired.get_ICER(),
      ICER_paired.get_CI(alpha=0.05, num_bootstrap_samples=1000),
      ICER_paired.get_CI(alpha=0.05, num_bootstrap_samples=1000, method='Bayesian'),
      ICER_paired.get_PI(alpha=0.05)))

# ICER calculation assuming independent observations
ICER_indp = EconEval.ICER_Indp('Testing independent ICER',
                               cost_intervention, effect_intervention, cost_base, effect_base,
                               health_measure='d')
print('Independent ICER (confidence and prediction interval): \n\t{}\n\t{}\n\t{}'.format(
      ICER_indp.get_ICER(),
      ICER_indp.get_CI(0.05, 1000),
      ICER_indp.get_PI(0.05)))

# try NMB
NMB_paired = EconEval.INMB_Paired("Testing paired NMB",
                                  cost_intervention, effect_intervention, cost_base, effect_base,
                                  health_measure='d')
print('Paired NMB (confidence and prediction interval): \n\t{}\n\t{}\n\t{}'.format(
      NMB_paired.get_INMB(wtp=10000),
      NMB_paired.get_CI(wtp=10000, alpha=.05),
      NMB_paired.get_PI(wtp=10000, alpha=.05)))

NMB_indp = EconEval.INMB_Indp("Testing independent NMB",
                              cost_intervention, effect_intervention, cost_base, effect_base,
                              health_measure='d')
print('Independent NMB (confidence and prediction interval): \n\t{}\n\t{}\n\t{}'.format(
      NMB_indp.get_INMB(wtp=10000),
      NMB_indp.get_CI(wtp=10000, alpha=.05),
      NMB_indp.get_PI(wtp=10000, alpha=.05)))

print('')



import numpy as np

import deampy.regression as Reg

# generate x values (observations)
x = np.random.randn(200)
# generate y values (assuming y = x^2 -2 + error)
y = pow(x, 2) - 4 * pow(x, 1) - 2 + np.random.randn(200)

polyReg = Reg.PolyRegression(x, y, degree=2)
polyReg.plot_fit(x_range=[-3, 3])

print(polyReg.get_coeffs())

print(polyReg.get_predicted_y(0))
print(polyReg.get_predicted_y(1))
print(polyReg.get_predicted_y(-1))

print(polyReg.get_derivative(0))
print(polyReg.get_derivative(1))

print(polyReg.get_roots())



import deampy.random_variats as RVGs
from deampy.models import MortalityModel

# from life tables
rows = [
    # age, sex , mortality rate
    [0,     0,  0.1],
    [0,     1,  0.125],
    [5,     0,  0.1],
    [5,     1,  0.25],
    [10,    0,  0.2],
    [10,    1,  0.225]
]

rng = RVGs.RNG(seed=1)

# mortality model
mortalityModel = MortalityModel(mortality_rates=rows,  # life table
                                group_mins=0,  # minimum value of sex group
                                group_maxs=1,  # maximum value of sex group
                                group_delta='int',  # sex group is a category
                                age_min=0,  # minimum age in this life table
                                age_delta=5)         # age interval

# get sample for time until death
print(mortalityModel.sample_time_to_death(group=0, age=8.9, rng=rng))
print(mortalityModel.sample_time_to_death(group=0, age=0, rng=rng))
print(mortalityModel.sample_time_to_death(group=0, age=32, rng=rng))





from math import pow


class _ExplorationRule:

    def get_epsilon(self, itr):
        pass


class _LearningRule:

    def get_step_size(self, itr):
        # goes to zero as itr goes to infinity
        return 0

    def get_forgetting_factor(self, itr):
        # goes to 1 as itr goes to infinity
        return 1/(1+self.get_step_size(itr))


class EpsilonGreedy(_ExplorationRule):
    # For selecting the greedy action with probability 1-epsilon.
    # epsilon_n = 1/n^beta, beta over (0.5, 1], n > 0

    def __init__(self, beta):
        self._beta = beta

    def __str__(self):
        return 'Beta{}'.format(self._beta)

    def get_epsilon(self, itr):

        return pow(itr, -self._beta)


class Harmonic(_LearningRule):
    # step_n = b / (b + n), for n >= 0 and b >= 1
    # (i is the iteration of the optimization algorithm)

    def __init__(self, b):
        self._b = b

    def __str__(self):
        return 'b{}'.format(self._b)

    def get_step_size(self, itr):
        return self._b / (self._b + itr - 1)



from deampy.plots.eff_sample_size import *

plot_eff_sample_size(likelihood_weights=[0.1, 0.2, 0.3, 0.4], if_randomize=False)
plot_eff_sample_size(likelihood_weights=[0.1, 0.2, 0.3, 0.4], if_randomize=True)



from deampy.sensitivity_analysis import SensitivityAnalysis


def do_sa(dic_parameter_values, outputs):
    sa = SensitivityAnalysis(dic_parameter_values=dic_parameter_values,
                             output_values=outputs)

    print('')
    sa.print_corr(corr='r')     # Pearson's
    sa.print_corr(corr='rho')   # Spearman's
    sa.print_corr(corr='p')     # partial correlation
    sa.print_corr(corr='pr')    # partial rank correlation


dic_parameter_values = {
    'par1': [1, 4, 3, 7, 1],
    'par2': [5, 7, 2, 8, 23],
    'par3': [5, 6, 7, 8, 9]
}
outputs = [7, 1, 2, 15, 5]

do_sa(dic_parameter_values=dic_parameter_values, outputs=outputs)


dic_parameter_values = {
    'par1': [1, 1, 1, 1, 1],
    'par2': [5, 7, 2, 8, 23],
}
outputs = [7, 1, 2, 15, 5]

do_sa(dic_parameter_values=dic_parameter_values, outputs=outputs)



import deampy.econ_eval as Econ


print('Present value of $10 collected in year 20 at discount rate 5%:')
print(Econ.pv_single_payment(payment=10, discount_rate=0.05, discount_period=20))

print('\nPresent value of $10 collected in year 20 at discount rate 5% (discounted continuously):')
print('These 2 numbers should be almost the same:')
print(Econ.pv_single_payment(payment=10, discount_rate=0.05,
                             discount_period=20, discount_continuously=True))
print(Econ.pv_single_payment(payment=10, discount_rate=0.05 / 100,
                             discount_period=20 * 100, discount_continuously=False))

print('\nPresent value of a continuous payment of $10 over the period [10, 20] at discount rate 5%:')
print(Econ.pv_continuous_payment(payment=10, discount_rate=0.05, discount_period=(10, 20)))
print('\nPresent value of a continuous payment of $10 over the period [10, 20] at discount rate 0%:')
print(Econ.pv_continuous_payment(payment=10, discount_rate=0, discount_period=(10, 20)))

print('\nEquivalent annual value of $50 over 10 years at discount rate 5%:')
print(Econ.equivalent_annual_value(present_value=50, discount_rate=0.05, discount_period=10))



import deampy.data_structure as df
import deampy.random_variats as RVGs

# one dimension dataframe for a continuous variable
df1 = df.OneDimDataFrame(y_objects=[10, 9, 20, 40], x_min=0, x_max=3, x_delta=1)

print('Testing one dimensional data frame with a continuous variable:')
print('print 10: ', df1.get_obj(x_value=0))
print('print 10: ', df1.get_obj(x_value=0.5))
print('print 9: ', df1.get_obj(x_value=1))
print('print 40: ', df1.get_obj(x_value=5))
print('print 40: ', df1.get_obj(x_value=3))
print('')

# one dimension dataframe for a categorical variable
df2 = df.OneDimDataFrame(y_objects=[10, 9, 20, 40], x_min=0, x_max=3, x_delta='int')

print('Testing one dimensional data frame with a categorical variable:')
print('print 10: ', df2.get_obj(x_value=0))
print('print 9: ', df2.get_obj(x_value=1))
print('')

# one dimension dataframe for a continuous variable
# assuming that y_values are rates of exponential distributions
df1 = df.OneDimDataFrameWithExpDist(y_values=[10, 9, 20, 40], x_min=0, x_max=3, x_delta=1)

rng = RVGs.RNG(seed=1)
print('Testing one dimensional data frame with exponential distributions')
print('Sample 1: ', df1.get_dist(x_value=0).sample(rng))
print('Sample 2: ', df1.get_dist(x_value=0.5).sample(rng))
print('Sample 3: ', df1.get_dist(x_value=1).sample(rng))
print('Sample 4: ', df1.get_dist(x_value=5).sample(rng))
print('')



import deampy.random_variats as rndSupport

from tests.prob_dists import rvg_tests as Tests

# use numpy random number generator
rng = rndSupport.RNG(1)

print('')

# tests
Tests.test_rng(rng)
Tests.test_bernoulli(rng, p=.2)
Tests.test_beta(rng, a=2, b=5, loc=1, scale=2)
Tests.test_beta(rng, a=0.2, b=0.1, loc=0.5, scale=0.25)
Tests.test_beta_binomial(rng, n=100, a=2, b=3, loc=1)
Tests.test_binomial(rng, n=1000, p=.2, loc=1)
Tests.test_dirichlet(rng, a=[1, 2, 3])
Tests.test_empirical(rng, prob=[0.2, 0.3, 0.5])
Tests.test_exponential(rng, scale=10, loc=1)
Tests.test_gamma(rng, a=2, scale=4, loc=1)
Tests.test_gamma_poisson(rng, a=2, gamma_scale=5, loc=1)
Tests.test_geometric(rng, p=.2, loc=1)
Tests.test_johnsonsb(rng, a=10, b=5, loc=10, scale=100)
Tests.test_johnsonsu(rng, a=10, b=3, loc=1, scale=2)
Tests.test_lognormal(rng, mu=0.2, sigma=0.1, loc=1)
Tests.test_multinomial(rng, n=100, pvals=[.2, 0.3, 0.5])
Tests.test_negative_binomial(rng, n=10, p=.2, loc=1)
Tests.test_non_homogeneous_exponential(rng, rates=[1, 1])
Tests.test_normal(rng, loc=5, scale=1.2)
Tests.test_poisson(rng, mu=2)
Tests.test_triangular(rng, c=0.2, loc=6, scale=7)
Tests.test_uniform(rng, loc=2, scale=7)
Tests.test_uniform_discrete(rng, l=1, u=5)
Tests.test_weibull(rng, a=0.5, loc=1, scale=1.5)



import numpy as np

from deampy.regression import LinearRegression, RecursiveLinearReg

coeffs = [1, -2, 3]  # y = 1 + 2x1 + 3x2
N = 1000
sigma = 1
l2_reg = 0.01
forgetting_factor = 0.95

np.random.seed(seed=1)

# generate X
X = []
for n in range(N):
    # 3 random numbers over [-1, 1]
    x = 2*np.random.sample(3)-1
    X.append(x)

# find y's
y = np.dot(X, np.array(coeffs)) + np.random.normal(0, sigma)

# fit a linear regression
lr = LinearRegression()
lr.fit(X=X, y=y)
print('Regression: ')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]) = ', lr.get_y(x=[1, -1, 2]))

# fit a linear regression with L2 regularization
lr = LinearRegression(l2_penalty=l2_reg)
lr.fit(X=X, y=y)
print('\nRegression (with l2-regularization):')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]) = ', lr.get_y(x=[1, -1, 2]))

# fit a linear regression with forgetting factor
lr = LinearRegression()
lr.fit(X=X, y=y, forgetting_factor=forgetting_factor)
print('\nRegression (with forgetting factor):')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]) = ', lr.get_y(x=[1, -1, 2]))

# recursive a linear regression
lr = RecursiveLinearReg()
for i in range(N):
    lr.update(x=X[i], y=y[i])
print('\nRecursive regression: ')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]', lr.get_y(x=[1, -1, 2]))

lr = RecursiveLinearReg(l2_penalty=l2_reg)
for i in range(N):
    lr.update(x=X[i], y=y[i])
print('\nRecursive regression (with l2-regularization): ')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]', lr.get_y(x=[1, -1, 2]))

lr = RecursiveLinearReg()
for i in range(N):
    lr.update(x=X[i], y=y[i], forgetting_factor=forgetting_factor)
print('\nRecursive regression (with forgetting factor): ')
print('Coeffs: ', lr.get_coeffs(), 'vs.', coeffs)
print('y([1, -1, 2]', lr.get_y(x=[1, -1, 2]))
